from functools import partial
from typing import (Callable,
                    Generic,
                    Optional,
                    cast)

from dendroid import red_black
from dendroid.hints import Key
from reprit.base import generate_repr
from robust.angular import (Orientation,
                            orientation)

from .event import (BinaryEvent,
                    Event,
                    NaryEvent)


class SweepLine(Generic[Event]):
    __slots__ = '_tree',

    def __init__(self, key: Callable[[Event], Key]) -> None:
        self._tree = red_black.set_(key=key)

    __repr__ = generate_repr(__init__)

    def __contains__(self, event: Event) -> bool:
        return event in self._tree

    def add(self, event: Event) -> None:
        self._tree.add(event)

    def remove(self, event: Event) -> None:
        self._tree.remove(event)

    def above(self, event: Event) -> Optional[Event]:
        try:
            return self._tree.next(event)
        except ValueError:
            return None

    def below(self, event: Event) -> Optional[Event]:
        try:
            return self._tree.prev(event)
        except ValueError:
            return None


class BinarySweepLineKey:
    __slots__ = 'event',

    def __init__(self, event: BinaryEvent) -> None:
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'BinarySweepLineKey') -> bool:
        """
        Checks if the segment (or at least the point) associated with event
        is lower than other's.
        """
        event, other_event = self.event, other.event
        if event is other_event:
            return False
        start, other_start = event.start, other_event.start
        end, other_end = event.end, other_event.end
        other_start_orientation = orientation(end, start, other_start)
        other_end_orientation = orientation(end, start, other_end)
        if other_start_orientation is other_end_orientation:
            start_x, start_y = start
            other_start_x, other_start_y = other_start
            if other_start_orientation is not Orientation.COLLINEAR:
                # other segment fully lies on one side
                return other_start_orientation is Orientation.COUNTERCLOCKWISE
            # segments are collinear
            elif event.from_left is not other_event.from_left:
                return event.from_left
            elif start_x == other_start_x:
                end_x, end_y = end
                other_end_x, other_end_y = other_end
                if start_y != other_start_y:
                    # segments are vertical
                    return start_y < other_start_y
                # segments have same start
                elif end_y != other_end_y:
                    return end_y < other_end_y
                else:
                    # segments are horizontal
                    return end_x < other_end_x
            elif start_y != other_start_y:
                return start_y < other_start_y
            else:
                # segments are horizontal
                return start_x < other_start_x
        start_orientation = orientation(other_end, other_start, start)
        end_orientation = orientation(other_end, other_start, end)
        if start_orientation is end_orientation:
            return start_orientation is Orientation.CLOCKWISE
        elif other_start_orientation is Orientation.COLLINEAR:
            return other_end_orientation is Orientation.COUNTERCLOCKWISE
        elif start_orientation is Orientation.COLLINEAR:
            return end_orientation is Orientation.CLOCKWISE
        elif end_orientation is Orientation.COLLINEAR:
            return start_orientation is Orientation.CLOCKWISE
        else:
            return other_start_orientation is Orientation.COUNTERCLOCKWISE


class NarySweepLineKey:
    __slots__ = 'event',

    def __init__(self, event: NaryEvent) -> None:
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'NarySweepLineKey') -> bool:
        """
        Checks if the segment (or at least the point) associated with event
        is lower than other's.
        """
        event, other_event = self.event, other.event
        if event is other_event:
            return False
        start, other_start = event.start, other_event.start
        end, other_end = event.end, other_event.end
        other_start_orientation = orientation(end, start, other_start)
        other_end_orientation = orientation(end, start, other_end)
        if other_start_orientation is other_end_orientation:
            start_x, start_y = start
            other_start_x, other_start_y = other_start
            if other_start_orientation is not Orientation.COLLINEAR:
                # other segment fully lies on one side
                return other_start_orientation is Orientation.COUNTERCLOCKWISE
            # segments are collinear
            elif start_x == other_start_x:
                end_x, end_y = end
                other_end_x, other_end_y = other_end
                if start_y != other_start_y:
                    # segments are vertical
                    return start_y < other_start_y
                # segments have same start
                elif end_y != other_end_y:
                    return end_y < other_end_y
                else:
                    # segments are horizontal
                    return end_x < other_end_x
            elif start_y != other_start_y:
                return start_y < other_start_y
            else:
                # segments are horizontal
                return start_x < other_start_x
        start_orientation = orientation(other_end, other_start, start)
        end_orientation = orientation(other_end, other_start, end)
        if start_orientation is end_orientation:
            return start_orientation is Orientation.CLOCKWISE
        elif other_start_orientation is Orientation.COLLINEAR:
            return other_end_orientation is Orientation.COUNTERCLOCKWISE
        elif start_orientation is Orientation.COLLINEAR:
            return end_orientation is Orientation.CLOCKWISE
        elif end_orientation is Orientation.COLLINEAR:
            return start_orientation is Orientation.CLOCKWISE
        else:
            return other_start_orientation is Orientation.COUNTERCLOCKWISE


BinarySweepLine = cast(Callable[[], SweepLine[BinaryEvent]],
                       partial(SweepLine, BinarySweepLineKey))
NarySweepLine = cast(Callable[[], SweepLine[NaryEvent]],
                     partial(SweepLine, NarySweepLineKey))



from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from numbers import Rational
from operator import attrgetter
from typing import (List,
                    Optional,
                    Union as Union_)

from reprit.base import generate_repr

from clipping.hints import (HolelessMix,
                            Multipoint,
                            Multiregion,
                            Multisegment)
from . import bounding_box
from .event import (ShapedEvent as Event,
                    events_to_connectivity)
from .events_queue import (BinaryEventsQueueKey as EventsQueueKey,
                           HolelessEventsQueue as EventsQueue)
from .sweep_line import BinarySweepLine as SweepLine
from .utils import (all_equal,
                    contour_to_oriented_segments,
                    pairwise,
                    to_multiregion_base,
                    to_multiregion_x_max,
                    to_rational_multiregion)


class Operation(ABC):
    __slots__ = 'left', 'right', 'accurate', '_events_queue'

    def __init__(self,
                 left: Multiregion,
                 right: Multiregion,
                 accurate: bool) -> None:
        """
        Initializes operation.

        :param left: left operand.
        :param right: right operand.
        :param accurate:
            flag that tells whether to use slow but more accurate arithmetic
            for floating point numbers.
        """
        self.left = left
        self.right = right
        self.accurate = accurate
        self._events_queue = EventsQueue()

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union_[Multiregion, HolelessMix]:
        """
        Computes result of the operation.
        """

    def compute_fields(self, event: Event, below_event: Optional[Event]
                       ) -> None:
        if below_event is not None:
            event.other_interior_to_left = (below_event.other_interior_to_left
                                            if (event.from_left
                                                is below_event.from_left)
                                            else below_event.interior_to_left)
        event.in_result = self.in_result(event)

    def fill_queue(self) -> None:
        events_queue = self._events_queue
        for region in self.left:
            events_queue.register_segments(
                    contour_to_oriented_segments(region), True)
        for region in self.right:
            events_queue.register_segments(
                    contour_to_oriented_segments(region), False)

    @abstractmethod
    def in_result(self, event: Event) -> bool:
        """Detects if event will be presented in result of the operation."""

    def normalize_operands(self) -> None:
        left, right = self.left, self.right
        if self.accurate and not issubclass(to_multiregion_base(left + right),
                                            Rational):
            self.left, self.right = (to_rational_multiregion(left),
                                     to_rational_multiregion(right))

    def process_event(self, event: Event, processed_events: List[Event],
                      sweep_line: SweepLine) -> None:
        if event.is_right_endpoint:
            processed_events.append(event)
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)
        elif event not in sweep_line:
            processed_events.append(event)
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            self.compute_fields(event, below_event)
            if (above_event is not None
                    and self._events_queue.detect_intersection(event,
                                                               above_event)):
                self.compute_fields(event, below_event)
                self.compute_fields(above_event, event)
            if (below_event is not None
                    and self._events_queue.detect_intersection(below_event,
                                                               event)):
                below_below_event = sweep_line.below(below_event)
                self.compute_fields(below_event, below_below_event)
                self.compute_fields(event, below_event)


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> HolelessMix:
        if not (self.left and self.right):
            return [], [], []
        left_bounding_box = bounding_box.from_multiregion(self.left)
        right_bounding_box = bounding_box.from_multiregion(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return [], [], []
        self.left = bounding_box.to_intersecting_regions(right_bounding_box,
                                                         self.left)
        self.right = bounding_box.to_intersecting_regions(left_bounding_box,
                                                          self.right)
        if not (self.left and self.right):
            return [], [], []
        self.normalize_operands()
        events = sorted(self.sweep(),
                        key=EventsQueueKey)
        multipoint = []  # type: Multipoint
        multisegment = []  # type: Multisegment
        for start, same_start_events in groupby(events,
                                                key=attrgetter('start')):
            same_start_events = list(same_start_events)
            if (all(event.is_right_endpoint or not event.in_result
                    for event in same_start_events)
                    and not all_equal(event.from_left
                                      for event in same_start_events)):
                no_segment_found = True
                for event, next_event in pairwise(same_start_events):
                    if (event.from_left is not next_event.from_left
                            and event.segment == next_event.segment):
                        no_segment_found = False
                        if not event.is_right_endpoint:
                            multisegment.append(next_event.segment)
                if no_segment_found and all(not event.complement.in_result
                                            if event.is_right_endpoint
                                            else not event.in_result
                                            for event in same_start_events):
                    multipoint.append(start)
        return multipoint, multisegment, events_to_multiregion(events)

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        sweep_line = SweepLine()
        min_max_x = min(to_multiregion_x_max(self.left),
                        to_multiregion_x_max(self.right))
        while self._events_queue:
            event = self._events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, result, sweep_line)
        return result

    def in_result(self, event: Event) -> bool:
        return (event.inside
                or not event.from_left and event.is_common_region_boundary)


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Multiregion:
        if not (self.left and self.right):
            return []
        left_bounding_box = bounding_box.from_multiregion(self.left)
        right_bounding_box = bounding_box.from_multiregion(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return []
        self.left = bounding_box.to_coupled_regions(right_bounding_box,
                                                    self.left)
        self.right = bounding_box.to_coupled_regions(left_bounding_box,
                                                     self.right)
        if not (self.left and self.right):
            return []
        self.normalize_operands()
        return events_to_multiregion(self.sweep())

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        sweep_line = SweepLine()
        min_max_x = min(to_multiregion_x_max(self.left),
                        to_multiregion_x_max(self.right))
        while self._events_queue:
            event = self._events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, result, sweep_line)
        return result

    def in_result(self, event: Event) -> bool:
        return (event.inside
                or not event.from_left and event.is_common_region_boundary)


def events_to_multiregion(events: List[Event]) -> Multiregion:
    events = sorted([event for event in events if event.primary.in_result],
                    key=EventsQueueKey)
    for index, event in enumerate(events):
        event.position = index
    processed = [False] * len(events)
    result = []
    connectivity = events_to_connectivity(events)
    for index, event in enumerate(events):
        if processed[index]:
            continue
        contour_start = event.start
        contour = [contour_start]
        contour_events = [event]
        cursor = event
        complement_position = event.complement.position
        processed[index] = processed[complement_position] = True
        while cursor.end != contour_start:
            contour.append(cursor.end)
            position = _to_next_position(complement_position, processed,
                                         connectivity)
            if position is None:
                break
            cursor = events[position]
            contour_events.append(cursor)
            complement_position = cursor.complement.position
            processed[position] = processed[complement_position] = True
        result.append(contour)
    return result


def _to_next_position(position: int,
                      processed: List[bool],
                      connectivity: List[int]) -> Optional[int]:
    candidate = position
    while True:
        candidate = connectivity[candidate]
        if not processed[candidate]:
            return candidate
        elif candidate == position:
            return None



from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from numbers import Rational
from operator import attrgetter
from typing import (List,
                    Optional,
                    Union as Union_)

from reprit.base import generate_repr

from clipping.hints import (Contour,
                            Mix,
                            Multipoint,
                            Multipolygon,
                            Multisegment)
from . import bounding_box
from .event import (HoleyEvent as Event,
                    events_to_connectivity)
from .events_queue import (BinaryEventsQueueKey as EventsQueueKey,
                           HoleyEventsQueue as EventsQueue)
from .sweep_line import BinarySweepLine as SweepLine
from .utils import (all_equal,
                    pairwise,
                    polygon_to_oriented_segments,
                    shrink_collinear_vertices,
                    to_first_boundary_vertex,
                    to_multipolygon_base,
                    to_multipolygon_x_max,
                    to_rational_multipolygon)


class Operation(ABC):
    __slots__ = 'left', 'right', 'accurate', '_events_queue'

    def __init__(self,
                 left: Multipolygon,
                 right: Multipolygon,
                 accurate: bool) -> None:
        """
        Initializes operation.

        :param left: left operand.
        :param right: right operand.
        :param accurate:
            flag that tells whether to use slow but more accurate arithmetic
            for floating point numbers.
        """
        self.left = left
        self.right = right
        self.accurate = accurate
        self._events_queue = EventsQueue()

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union_[Multipolygon, Mix]:
        """
        Computes result of the operation.
        """

    def compute_fields(self, event: Event, below_event: Optional[Event]
                       ) -> None:
        if below_event is not None:
            event.other_interior_to_left = (below_event.other_interior_to_left
                                            if (event.from_left
                                                is below_event.from_left)
                                            else below_event.interior_to_left)
            event.below_in_result_event = (below_event.below_in_result_event
                                           if (not self.in_result(below_event)
                                               or below_event.is_vertical)
                                           else below_event)
        event.in_result = self.in_result(event)

    def fill_queue(self) -> None:
        events_queue = self._events_queue
        for polygon in self.left:
            events_queue.register_segments(
                    polygon_to_oriented_segments(polygon), True)
        for polygon in self.right:
            events_queue.register_segments(
                    polygon_to_oriented_segments(polygon), False)

    @abstractmethod
    def in_result(self, event: Event) -> bool:
        """Detects if event will be presented in result of the operation."""

    def normalize_operands(self) -> None:
        left, right = self.left, self.right
        if (self.accurate
                and not issubclass(to_multipolygon_base(left + right),
                                   Rational)):
            self.left, self.right = (to_rational_multipolygon(left),
                                     to_rational_multipolygon(right))

    def process_event(self, event: Event, processed_events: List[Event],
                      sweep_line: SweepLine) -> None:
        if event.is_right_endpoint:
            processed_events.append(event)
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)
        elif event not in sweep_line:
            processed_events.append(event)
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            self.compute_fields(event, below_event)
            if (above_event is not None
                    and self._events_queue.detect_intersection(event,
                                                               above_event)):
                self.compute_fields(event, below_event)
                self.compute_fields(above_event, event)
            if (below_event is not None
                    and self._events_queue.detect_intersection(below_event,
                                                               event)):
                below_below_event = sweep_line.below(below_event)
                self.compute_fields(below_event, below_below_event)
                self.compute_fields(event, below_event)

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        sweep_line = SweepLine()
        events_queue = self._events_queue
        while events_queue:
            self.process_event(events_queue.pop(), result, sweep_line)
        return result


class Difference(Operation):
    __slots__ = ()

    def compute(self) -> Multipolygon:
        if not (self.left and self.right):
            return self.left
        left_bounding_box = bounding_box.from_multipolygon(self.left)
        if bounding_box.disjoint_with(
                left_bounding_box, bounding_box.from_multipolygon(self.right)):
            return self.left
        self.right = bounding_box.to_coupled_polygons(left_bounding_box,
                                                      self.right)
        if not self.right:
            return self.left
        self.normalize_operands()
        return events_to_multipolygon(self.sweep())

    def in_result(self, event: Event) -> bool:
        return (event.outside
                if event.from_left
                else event.inside or event.is_common_polyline_component)

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        left_x_max = to_multipolygon_x_max(self.left)
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > left_x_max:
                break
            self.process_event(event, result, sweep_line)
        return result


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> Mix:
        if not (self.left and self.right):
            return [], [], []
        left_bounding_box = bounding_box.from_multipolygon(self.left)
        right_bounding_box = bounding_box.from_multipolygon(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return [], [], []
        self.left = bounding_box.to_intersecting_polygons(right_bounding_box,
                                                          self.left)
        self.right = bounding_box.to_intersecting_polygons(left_bounding_box,
                                                           self.right)
        if not (self.left and self.right):
            return [], [], []
        self.normalize_operands()
        events = sorted(self.sweep(),
                        key=EventsQueueKey)
        multipoint = []  # type: Multipoint
        multisegment = []  # type: Multisegment
        for start, same_start_events in groupby(events,
                                                key=attrgetter('start')):
            same_start_events = list(same_start_events)
            if (all(event.is_right_endpoint or not event.in_result
                    for event in same_start_events)
                    and not all_equal(event.from_left
                                      for event in same_start_events)):
                no_segment_found = True
                for event, next_event in pairwise(same_start_events):
                    if (event.from_left is not next_event.from_left
                            and event.segment == next_event.segment):
                        no_segment_found = False
                        if not event.is_right_endpoint:
                            multisegment.append(next_event.segment)
                if no_segment_found and all(not event.primary.in_result
                                            for event in same_start_events):
                    multipoint.append(start)
        return multipoint, multisegment, events_to_multipolygon(events)

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        min_max_x = min(to_multipolygon_x_max(self.left),
                        to_multipolygon_x_max(self.right))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, result, sweep_line)
        return result

    def in_result(self, event: Event) -> bool:
        return (event.inside
                or not event.from_left and event.is_common_region_boundary)


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Multipolygon:
        if not (self.left and self.right):
            return []
        left_bounding_box = bounding_box.from_multipolygon(self.left)
        right_bounding_box = bounding_box.from_multipolygon(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return []
        self.left = bounding_box.to_coupled_polygons(right_bounding_box,
                                                     self.left)
        self.right = bounding_box.to_coupled_polygons(left_bounding_box,
                                                      self.right)
        if not (self.left and self.right):
            return []
        self.normalize_operands()
        return events_to_multipolygon(self.sweep())

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        min_max_x = min(to_multipolygon_x_max(self.left),
                        to_multipolygon_x_max(self.right))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, result, sweep_line)
        return result

    def in_result(self, event: Event) -> bool:
        return (event.inside
                or not event.from_left and event.is_common_region_boundary)


class SymmetricDifference(Operation):
    __slots__ = ()

    def compute(self) -> Multipolygon:
        if not (self.left and self.right):
            return self.left or self.right
        elif bounding_box.disjoint_with(
                bounding_box.from_multipolygon(self.left),
                bounding_box.from_multipolygon(self.right)):
            result = self.left + self.right
            result.sort(key=to_first_boundary_vertex)
            return result
        self.normalize_operands()
        return events_to_multipolygon(self.sweep())

    def in_result(self, event: Event) -> bool:
        return not event.is_overlap


class Union(Operation):
    __slots__ = ()

    def compute(self) -> Multipolygon:
        if not (self.left and self.right):
            return self.left or self.right
        elif bounding_box.disjoint_with(
                bounding_box.from_multipolygon(self.left),
                bounding_box.from_multipolygon(self.right)):
            result = self.left + self.right
            result.sort(key=to_first_boundary_vertex)
            return result
        self.normalize_operands()
        return events_to_multipolygon(self.sweep())

    def in_result(self, event: Event) -> bool:
        return (event.outside
                or not event.from_left and event.is_common_region_boundary)


def events_to_multipolygon(events: List[Event]) -> Multipolygon:
    events = sorted([event for event in events if event.primary.in_result],
                    key=EventsQueueKey)
    for index, event in enumerate(events):
        event.position = index
    are_internal, depths, holes, parents = [], [], [], []
    processed = [False] * len(events)
    contours = []
    connectivity = events_to_connectivity(events)
    for index, event in enumerate(events):
        if processed[index]:
            continue
        contour_id = len(contours)
        _compute_relations(event, contour_id, are_internal, depths, holes,
                           parents)
        contour = _events_to_contour(event, events, contour_id, connectivity,
                                     processed)
        shrink_collinear_vertices(contour)
        if depths[contour_id] % 2:
            # holes will be in clockwise order
            contour.reverse()
        contours.append(contour)
    result = []
    for index, contour in enumerate(contours):
        if are_internal[index]:
            # hole of a hole is an external polygon
            result.extend((contours[hole_index],
                           [contours[hole_hole_index]
                            for hole_hole_index in holes[hole_index]])
                          for hole_index in holes[index])
        else:
            result.append((contour, [contours[hole_index]
                                     for hole_index in holes[index]]))
    return result


def _compute_relations(event: Event,
                       contour_id: int,
                       are_internal: List[bool],
                       depths: List[int],
                       holes: List[List[int]],
                       parents: List[Optional[int]]) -> None:
    depth = 0
    parent = None
    is_internal = False
    below_in_result_event = event.below_in_result_event
    if below_in_result_event is not None:
        below_in_result_contour_id = below_in_result_event.contour_id
        if not below_in_result_event.result_in_out:
            holes[below_in_result_contour_id].append(contour_id)
            parent = below_in_result_contour_id
            depth = depths[below_in_result_contour_id] + 1
            is_internal = True
        elif are_internal[below_in_result_contour_id]:
            below_in_result_parent_id = parents[below_in_result_contour_id]
            holes[below_in_result_parent_id].append(contour_id)
            parent = below_in_result_parent_id
            depth = depths[below_in_result_contour_id]
            is_internal = True
    holes.append([])
    parents.append(parent)
    depths.append(depth)
    are_internal.append(is_internal)


def _events_to_contour(cursor: Event,
                       events: List[Event],
                       contour_id: int,
                       connectivity: List[int],
                       processed: List[bool]) -> Contour:
    contour_start = cursor.start
    contour = [contour_start]
    contour_events = [cursor]
    complement_position = cursor.complement.position
    vertices_positions = {contour_start: 0}
    while cursor.end != contour_start:
        vertex = cursor.end
        if vertex in vertices_positions:
            # vertices loop found, i.e. contour has self-intersection
            previous_vertex_position = vertices_positions[vertex]
            del contour[previous_vertex_position:]
            del contour_events[previous_vertex_position:]
        else:
            vertices_positions[vertex] = len(contour)
        contour.append(vertex)
        position = _to_next_position(complement_position, processed,
                                     connectivity)
        if position is None:
            break
        cursor = events[position]
        contour_events.append(cursor)
        complement_position = cursor.complement.position
    for event in contour_events:
        processed[event.position] = processed[event.complement.position] = True
        if event.is_right_endpoint:
            event.complement.result_in_out = True
            event.complement.contour_id = contour_id
        else:
            event.result_in_out = False
            event.contour_id = contour_id
    return contour


def _to_next_position(position: int,
                      processed: List[bool],
                      connectivity: List[int]) -> Optional[int]:
    candidate = position
    while True:
        candidate = connectivity[candidate]
        if not processed[candidate]:
            return candidate
        elif candidate == position:
            return None



"""
Boolean operations on multisegments/multiregions/multipolygons in the plane.

Based on algorithm by F. Martinez et al.

Reference:
    https://doi.org/10.1016/j.advengsoft.2013.04.004
    http://www4.ujaen.es/~fmartin/bool_op.html

########
Glossary
########

**Point** --- pair of real numbers (called *point's coordinates*).

**Multipoint** --- possibly empty sequence of distinct points.

**Segment** (or **line segment**) --- pair of unequal points
(called *segment's endpoints*).

**Multisegment** --- possibly empty sequence of segments
such that any pair of them do not cross/overlap each other.

**Contour** --- sequence of points (called *contour's vertices*)
such that line segments formed by pairs of consecutive points
(including the last-first point pair)
do not overlap each other.

**Region** --- contour with points that lie within it.

**Multiregion** --- possibly empty sequence of regions such
that intersection of distinct regions is a discrete points set.

**Polygon** --- pair of region (called *polygon's border*)
and multiregion which lies within the border (called *polygon's holes*).

**Multipolygon** --- possibly empty sequence of non-overlapping polygons.

**Mix** --- triplet of disjoint/touching multipoint, multisegment
and multipolygon.
"""
from itertools import groupby as _groupby
from typing import List

from .core import (holeless as _holeless,
                   holey as _holey,
                   linear as _linear,
                   mixed as _mixed)
from .hints import (HolelessMix,
                    Mix,
                    Multipolygon,
                    Multiregion,
                    Multisegment,
                    Segment)


def segments_to_multisegment(segments: List[Segment],
                             *,
                             accurate: bool = True) -> Multisegment:
    """
    Returns multisegment from given segments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(segments)``,
    ``intersections_count`` --- number of intersections between segments.

    :param segments: target segments.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: multisegment from segments.

    >>> segments_to_multisegment([])
    []
    >>> segments_to_multisegment([((0, 0), (1, 0)), ((0, 1), (1, 0))])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> segments_to_multisegment([((0, 0), (2, 0)), ((1, 0), (3, 0))])
    [((0, 0), (1, 0)), ((1, 0), (2, 0)), ((2, 0), (3, 0))]
    """
    return sorted(segment
                  for segment, _ in _groupby(_linear.merge_segments(segments,
                                                                    accurate)))


def complete_intersect_multisegments(left: Multisegment,
                                     right: Multisegment,
                                     *,
                                     accurate: bool = True) -> Mix:
    """
    Returns intersection of multisegments considering cases
    with segments touching each other in points.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(left) + len(right)``,
    ``intersections_count`` --- number of intersections between multisegments.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> complete_intersect_multisegments([], [])
    ([], [], [])
    >>> complete_intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                                  [])
    ([], [], [])
    >>> complete_intersect_multisegments([],
    ...                                  [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    ([], [], [])
    >>> complete_intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                                  [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    ([], [((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    >>> complete_intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 1))],
    ...                                  [((0, 0), (2, 0)), ((0, 0), (2, 2))])
    ([(1, 1)], [((0, 0), (1, 0))], [])
    """
    return _linear.CompleteIntersection(left, right, accurate).compute()


def intersect_multisegments(left: Multisegment,
                            right: Multisegment,
                            *,
                            accurate: bool = True) -> Multisegment:
    """
    Returns intersection of multisegments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(left) + len(right)``,
    ``intersections_count`` --- number of intersections between multisegments.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> intersect_multisegments([], [])
    []
    >>> intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    []
    >>> intersect_multisegments([], [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    []
    >>> intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                         [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> intersect_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 1))],
    ...                         [((0, 0), (2, 0)), ((0, 0), (2, 2))])
    [((0, 0), (1, 0))]
    """
    return _linear.Intersection(left, right, accurate).compute()


def subtract_multisegments(minuend: Multisegment,
                           subtrahend: Multisegment,
                           *,
                           accurate: bool = True) -> Multisegment:
    """
    Returns difference of multisegments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(left) + len(right)``,
    ``intersections_count`` --- number of intersections between multisegments.

    :param minuend: multisegment to subtract from.
    :param subtrahend: multisegment to subtract.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: difference between minuend and subtrahend.

    >>> subtract_multisegments([], [])
    []
    >>> subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> subtract_multisegments([], [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    []
    >>> subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                        [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    []
    >>> subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 1))],
    ...                        [((0, 0), (2, 0)), ((0, 0), (2, 2))])
    [((0, 1), (1, 1))]
    """
    return _linear.Difference(minuend, subtrahend, accurate).compute()


def symmetric_subtract_multisegments(left: Multisegment,
                                     right: Multisegment,
                                     *,
                                     accurate: bool = True) -> Multisegment:
    """
    Returns symmetric difference of multisegments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(left) + len(right)``,
    ``intersections_count`` --- number of intersections between multisegments.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: symmetric difference of operands.

    >>> symmetric_subtract_multisegments([], [])
    []
    >>> symmetric_subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                                  [])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> symmetric_subtract_multisegments([],
    ...                                  [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> symmetric_subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                                  [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    []
    >>> symmetric_subtract_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 1))],
    ...                                  [((0, 0), (2, 0)), ((0, 0), (2, 2))])
    [((0, 0), (1, 1)), ((0, 1), (1, 1)), ((1, 0), (2, 0)), ((1, 1), (2, 2))]
    """
    return _linear.SymmetricDifference(left, right, accurate).compute()


def unite_multisegments(left: Multisegment,
                        right: Multisegment,
                        *,
                        accurate: bool = True) -> Multisegment:
    """
    Returns union of multisegments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = segments_count + intersections_count``,
    ``segments_count = len(left) + len(right)``,
    ``intersections_count`` --- number of intersections between multisegments.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: union of operands.

    >>> unite_multisegments([], [])
    []
    >>> unite_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> unite_multisegments([], [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> unite_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...                     [((0, 0), (1, 0)), ((0, 1), (1, 0))])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> unite_multisegments([((0, 0), (1, 0)), ((0, 1), (1, 1))],
    ...                     [((0, 0), (2, 0)), ((0, 0), (2, 2))])
    [((0, 0), (1, 0)), ((0, 0), (1, 1)), ((0, 1), (1, 1)), ((1, 0), (2, 0)),\
 ((1, 1), (2, 2))]
    """
    return _linear.Union(left, right, accurate).compute()


def intersect_multisegment_with_multipolygon(multisegment: Multisegment,
                                             multipolygon: Multipolygon,
                                             *,
                                             accurate: bool = True
                                             ) -> Multisegment:
    """
    Returns intersection of multisegment with multipolygon.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = start_segments_count + intersections_count``,
    ``start_segments_count = len(multisegment) + multipolygon_edges_count``,
    ``multipolygon_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in multipolygon)``,
    ``intersections_count`` --- number of intersections between multisegment
    and multipolygon edges.

    :param multisegment: multisegment to intersect with.
    :param multipolygon: multipolygon to intersect with.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of multisegment with multipolygon.

    >>> intersect_multisegment_with_multipolygon([], [])
    []
    >>> intersect_multisegment_with_multipolygon(
    ...         [], [([(0, 0), (1, 0), (0, 1)], [])])
    []
    >>> intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    []
    >>> intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...         [([(0, 0), (1, 0), (0, 1)], [])])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((1, 1), (2, 2))],
    ...         [([(0, 0), (1, 0), (1, 1), (0, 1)], [])])
    [((0, 0), (1, 0))]
    """
    return _mixed.Intersection(multisegment, multipolygon, accurate).compute()


def complete_intersect_multisegment_with_multipolygon(
        multisegment: Multisegment,
        multipolygon: Multipolygon,
        *,
        accurate: bool = True) -> Mix:
    """
    Returns intersection of multisegment with multipolygon considering cases
    with geometries touching each other in points/segments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = start_segments_count + intersections_count``,
    ``start_segments_count = len(multisegment) + multipolygon_edges_count``,
    ``multipolygon_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in multipolygon)``,
    ``intersections_count`` --- number of intersections between multisegment
    and multipolygon edges.

    :param multisegment: multisegment to intersect with.
    :param multipolygon: multipolygon to intersect with.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of multisegment with multipolygon.

    >>> complete_intersect_multisegment_with_multipolygon([], [])
    ([], [], [])
    >>> complete_intersect_multisegment_with_multipolygon(
    ...         [], [([(0, 0), (1, 0), (0, 1)], [])])
    ([], [], [])
    >>> complete_intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    ([], [], [])
    >>> complete_intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...         [([(0, 0), (1, 0), (0, 1)], [])])
    ([], [((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    >>> complete_intersect_multisegment_with_multipolygon(
    ...         [((0, 0), (1, 0)), ((1, 1), (2, 2))],
    ...         [([(0, 0), (1, 0), (1, 1), (0, 1)], [])])
    ([(1, 1)], [((0, 0), (1, 0))], [])
    """
    return _mixed.CompleteIntersection(multisegment, multipolygon,
                                       accurate).compute()


def subtract_multipolygon_from_multisegment(multisegment: Multisegment,
                                            multipolygon: Multipolygon,
                                            *,
                                            accurate: bool = True
                                            ) -> Multisegment:
    """
    Returns difference of multisegment with multipolygon.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = start_segments_count + intersections_count``,
    ``start_segments_count = len(multisegment) + multipolygon_edges_count``,
    ``multipolygon_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in multipolygon)``,
    ``intersections_count`` --- number of intersections between multisegment
    and multipolygon edges.

    :param multisegment: multisegment to subtract from.
    :param multipolygon: multipolygon to subtract.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: difference of multisegment with multipolygon.

    >>> subtract_multipolygon_from_multisegment([], [])
    []
    >>> subtract_multipolygon_from_multisegment(
    ...         [], [([(0, 0), (1, 0), (0, 1)], [])])
    []
    >>> subtract_multipolygon_from_multisegment(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))], [])
    [((0, 0), (1, 0)), ((0, 1), (1, 0))]
    >>> subtract_multipolygon_from_multisegment(
    ...         [((0, 0), (1, 0)), ((0, 1), (1, 0))],
    ...         [([(0, 0), (1, 0), (0, 1)], [])])
    []
    >>> subtract_multipolygon_from_multisegment(
    ...         [((0, 0), (1, 0)), ((1, 1), (2, 2))],
    ...         [([(0, 0), (1, 0), (1, 1), (0, 1)], [])])
    [((1, 1), (2, 2))]
    """
    return _mixed.Difference(multisegment, multipolygon, accurate).compute()


def complete_intersect_multiregions(left: Multiregion,
                                    right: Multiregion,
                                    *,
                                    accurate: bool = True) -> HolelessMix:
    """
    Returns intersection of multiregions considering cases
    with regions touching each other in points/segments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(map(len, left))``,
    ``right_edges_count = sum(map(len, right))``,
    ``intersections_count`` --- number of intersections between multiregions
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> lower_left_square = [(0, 0), (1, 0), (1, 1), (0, 1)]
    >>> lower_right_square = [(1, 0), (2, 0), (2, 1), (1, 1)]
    >>> upper_left_square = [(0, 1), (1, 1), (1, 2), (0, 2)]
    >>> upper_right_square = [(1, 1), (2, 1), (2, 2), (1, 2)]
    >>> complete_intersect_multiregions([], [])
    ([], [], [])
    >>> complete_intersect_multiregions([lower_left_square], [])
    ([], [], [])
    >>> complete_intersect_multiregions([], [lower_left_square])
    ([], [], [])
    >>> complete_intersect_multiregions([lower_left_square],
    ...                                 [lower_left_square])
    ([], [], [[(0, 0), (1, 0), (1, 1), (0, 1)]])
    >>> complete_intersect_multiregions([lower_left_square],
    ...                                 [lower_right_square])
    ([], [((1, 0), (1, 1))], [])
    >>> complete_intersect_multiregions([lower_left_square],
    ...                                 [upper_left_square])
    ([], [((0, 1), (1, 1))], [])
    >>> complete_intersect_multiregions([lower_left_square],
    ...                                 [upper_right_square])
    ([(1, 1)], [], [])
    >>> complete_intersect_multiregions([lower_left_square,
    ...                                  upper_right_square],
    ...                                 [upper_left_square,
    ...                                  lower_right_square])
    ([],\
 [((0, 1), (1, 1)), ((1, 0), (1, 1)), ((1, 1), (2, 1)), ((1, 1), (1, 2))], [])
    >>> complete_intersect_multiregions([lower_left_square,
    ...                                  upper_right_square],
    ...                                 [lower_left_square,
    ...                                  upper_right_square])
    ([], [],\
 [[(0, 0), (1, 0), (1, 1), (0, 1)], [(1, 1), (2, 1), (2, 2), (1, 2)]])
    """
    return _holeless.CompleteIntersection(left, right, accurate).compute()


def intersect_multiregions(left: Multiregion,
                           right: Multiregion,
                           *,
                           accurate: bool = True) -> Multiregion:
    """
    Returns intersection of multiregions.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(map(len, left))``,
    ``right_edges_count = sum(map(len, right))``,
    ``intersections_count`` --- number of intersections between multiregions
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> lower_left_square = [(0, 0), (1, 0), (1, 1), (0, 1)]
    >>> lower_right_square = [(1, 0), (2, 0), (2, 1), (1, 1)]
    >>> upper_left_square = [(0, 1), (1, 1), (1, 2), (0, 2)]
    >>> upper_right_square = [(1, 1), (2, 1), (2, 2), (1, 2)]
    >>> intersect_multiregions([], [])
    []
    >>> intersect_multiregions([lower_left_square], [])
    []
    >>> intersect_multiregions([], [lower_left_square])
    []
    >>> intersect_multiregions([lower_left_square], [lower_left_square])
    [[(0, 0), (1, 0), (1, 1), (0, 1)]]
    >>> intersect_multiregions([lower_left_square], [lower_right_square])
    []
    >>> intersect_multiregions([lower_left_square], [upper_left_square])
    []
    >>> intersect_multiregions([lower_left_square], [upper_right_square])
    []
    >>> intersect_multiregions([lower_left_square, upper_right_square],
    ...                        [upper_left_square, lower_right_square])
    []
    >>> intersect_multiregions([lower_left_square, upper_right_square],
    ...                        [lower_left_square, upper_right_square])
    [[(0, 0), (1, 0), (1, 1), (0, 1)], [(1, 1), (2, 1), (2, 2), (1, 2)]]
    """
    return _holeless.Intersection(left, right, accurate).compute()


def complete_intersect_multipolygons(left: Multipolygon,
                                     right: Multipolygon,
                                     *,
                                     accurate: bool = True) -> Mix:
    """
    Returns intersection of multipolygons considering cases
    with polygons touching each other in points/segments.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in left)``,
    ``right_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in right)``,
    ``intersections_count`` --- number of intersections between multipolygons
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> lower_left_square = [(0, 0), (3, 0), (3, 3), (0, 3)]
    >>> lower_left_triangle = [(2, 1), (2, 2), (1, 2)]
    >>> lower_right_square = [(3, 0), (6, 0), (6, 3), (3, 3)]
    >>> lower_right_triangle = [(4, 1), (5, 2), (4, 2)]
    >>> upper_left_square = [(0, 3), (3, 3), (3, 6), (0, 6)]
    >>> upper_left_triangle = [(1, 4), (2, 4), (2, 5)]
    >>> upper_right_square = [(3, 3), (6, 3), (6, 6), (3, 6)]
    >>> upper_right_triangle = [(4, 4), (5, 4), (4, 5)]
    >>> complete_intersect_multipolygons([], [])
    ([], [], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])], [])
    ([], [], [])
    >>> complete_intersect_multipolygons([], [(lower_left_square,
    ...                                        [lower_left_triangle])])
    ([], [], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(lower_left_square,
    ...                                    [lower_left_triangle])])
    ([], [], [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]])])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(lower_right_square,
    ...                                    [lower_right_triangle])])
    ([], [((3, 0), (3, 3))], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(upper_left_square,
    ...                                    [upper_left_triangle])])
    ([], [((0, 3), (3, 3))], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(upper_right_square,
    ...                                    [upper_right_triangle])])
    ([(3, 3)], [], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])],
    ...                                  [(upper_left_square,
    ...                                    [upper_left_triangle]),
    ...                                   (lower_right_square,
    ...                                    [lower_right_triangle])])
    ([],\
 [((0, 3), (3, 3)), ((3, 0), (3, 3)), ((3, 3), (6, 3)), ((3, 3), (3, 6))], [])
    >>> complete_intersect_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])],
    ...                                  [(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])])
    ([], [],\
 [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])])
    """
    return _holey.CompleteIntersection(left, right, accurate).compute()


def intersect_multipolygons(left: Multipolygon,
                            right: Multipolygon,
                            *,
                            accurate: bool = True) -> Multipolygon:
    """
    Returns intersection of multipolygons.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in left)``,
    ``right_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in right)``,
    ``intersections_count`` --- number of intersections between multipolygons
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: intersection of operands.

    >>> lower_left_square = [(0, 0), (3, 0), (3, 3), (0, 3)]
    >>> lower_left_triangle = [(2, 1), (2, 2), (1, 2)]
    >>> lower_right_square = [(3, 0), (6, 0), (6, 3), (3, 3)]
    >>> lower_right_triangle = [(4, 1), (5, 2), (4, 2)]
    >>> upper_left_square = [(0, 3), (3, 3), (3, 6), (0, 6)]
    >>> upper_left_triangle = [(1, 4), (2, 4), (2, 5)]
    >>> upper_right_square = [(3, 3), (6, 3), (6, 6), (3, 6)]
    >>> upper_right_triangle = [(4, 4), (5, 4), (4, 5)]
    >>> intersect_multipolygons([], [])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                         [])
    []
    >>> intersect_multipolygons([], [(lower_left_square,
    ...                               [lower_left_triangle])])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                         [(lower_left_square, [lower_left_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]])]
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                         [(lower_right_square, [lower_right_triangle])])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                         [(upper_left_square, [upper_left_triangle])])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                         [(upper_right_square, [upper_right_triangle])])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                          (upper_right_square, [upper_right_triangle])],
    ...                         [(upper_left_square, [upper_left_triangle]),
    ...                          (lower_right_square, [lower_right_triangle])])
    []
    >>> intersect_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                          (upper_right_square, [upper_right_triangle])],
    ...                         [(lower_left_square, [lower_left_triangle]),
    ...                          (upper_right_square, [upper_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])]
    """
    return _holey.Intersection(left, right, accurate).compute()


def subtract_multipolygons(minuend: Multipolygon,
                           subtrahend: Multipolygon,
                           *,
                           accurate: bool = True) -> Multipolygon:
    """
    Returns difference of multipolygons.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in left)``,
    ``right_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in right)``,
    ``intersections_count`` --- number of intersections between multipolygons
    edges.

    :param minuend: multipolygon to subtract from.
    :param subtrahend: multipolygon to subtract.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: difference between minuend and subtrahend.

    >>> lower_left_square = [(0, 0), (3, 0), (3, 3), (0, 3)]
    >>> lower_left_triangle = [(2, 1), (2, 2), (1, 2)]
    >>> lower_right_square = [(3, 0), (6, 0), (6, 3), (3, 3)]
    >>> lower_right_triangle = [(4, 1), (5, 2), (4, 2)]
    >>> upper_left_square = [(0, 3), (3, 3), (3, 6), (0, 6)]
    >>> upper_left_triangle = [(1, 4), (2, 4), (2, 5)]
    >>> upper_right_square = [(3, 3), (6, 3), (6, 6), (3, 6)]
    >>> upper_right_triangle = [(4, 4), (5, 4), (4, 5)]
    >>> subtract_multipolygons([], [])
    []
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                        [])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> subtract_multipolygons([], [(lower_left_square,
    ...                              [lower_left_triangle])])
    []
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                        [(lower_left_square, [lower_left_triangle])])
    []
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                        [(lower_right_square, [lower_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]])]
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                        [(upper_left_square, [upper_left_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]])]
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                        [(upper_right_square, [upper_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                         (upper_right_square, [upper_right_triangle])],
    ...                        [(upper_left_square, [upper_left_triangle]),
    ...                         (lower_right_square, [lower_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])]
    >>> subtract_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                         (upper_right_square, [upper_right_triangle])],
    ...                        [(lower_left_square, [lower_left_triangle]),
    ...                         (upper_right_square, [upper_right_triangle])])
    []
    """
    return _holey.Difference(minuend, subtrahend, accurate).compute()


def symmetric_subtract_multipolygons(left: Multipolygon,
                                     right: Multipolygon,
                                     *,
                                     accurate: bool = True) -> Multipolygon:
    """
    Returns symmetric difference of multipolygons.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in left)``,
    ``right_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in right)``,
    ``intersections_count`` --- number of intersections between multipolygons
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: symmetric difference of operands.

    >>> lower_left_square = [(0, 0), (3, 0), (3, 3), (0, 3)]
    >>> lower_left_triangle = [(2, 1), (2, 2), (1, 2)]
    >>> lower_right_square = [(3, 0), (6, 0), (6, 3), (3, 3)]
    >>> lower_right_triangle = [(4, 1), (5, 2), (4, 2)]
    >>> upper_left_square = [(0, 3), (3, 3), (3, 6), (0, 6)]
    >>> upper_left_triangle = [(1, 4), (2, 4), (2, 5)]
    >>> upper_right_square = [(3, 3), (6, 3), (6, 6), (3, 6)]
    >>> upper_right_triangle = [(4, 4), (5, 4), (4, 5)]
    >>> symmetric_subtract_multipolygons([], [])
    []
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                   [lower_left_triangle])], [])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> symmetric_subtract_multipolygons([], [(lower_left_square,
    ...                                       [lower_left_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(lower_left_square,
    ...                                    [lower_left_triangle])])
    []
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(lower_right_square,
    ...                                    [lower_right_triangle])])
    [([(0, 0), (6, 0), (6, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)],\
 [(4, 2), (5, 2), (4, 1)]])]
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(upper_left_square,
    ...                                    [upper_left_triangle])])
    [([(0, 0), (3, 0), (3, 6), (0, 6)], [[(2, 2), (2, 1), (1, 2)],\
 [(2, 5), (2, 4), (1, 4)]])]
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle])],
    ...                                  [(upper_right_square,
    ...                                    [upper_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])]
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])],
    ...                                  [(upper_left_square,
    ...                                    [upper_left_triangle]),
    ...                                   (lower_right_square,
    ...                                    [lower_right_triangle])])
    [([(0, 0), (6, 0), (6, 6), (0, 6)], [[(2, 2), (2, 1), (1, 2)],\
 [(2, 5), (2, 4), (1, 4)], [(4, 2), (5, 2), (4, 1)],\
 [(4, 5), (5, 4), (4, 4)]])]
    >>> symmetric_subtract_multipolygons([(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])],
    ...                                  [(lower_left_square,
    ...                                    [lower_left_triangle]),
    ...                                   (upper_right_square,
    ...                                    [upper_right_triangle])])
    []
    """
    return _holey.SymmetricDifference(left, right, accurate).compute()


def unite_multipolygons(left: Multipolygon,
                        right: Multipolygon,
                        *,
                        accurate: bool = True) -> Multipolygon:
    """
    Returns union of multipolygons.

    Time complexity:
        ``O(segments_count * log segments_count)``
    Memory complexity:
        ``O(segments_count)``

    where ``segments_count = edges_count + intersections_count``,
    ``edges_count = left_edges_count + right_edges_count``,
    ``left_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in left)``,
    ``right_edges_count = sum(len(border) + sum(map(len, holes))\
 for border, holes in right)``,
    ``intersections_count`` --- number of intersections between multipolygons
    edges.

    :param left: left operand.
    :param right: right operand.
    :param accurate:
        flag that tells whether to use slow but more accurate arithmetic
        for floating point numbers.
    :returns: union of operands.

    >>> lower_left_square = [(0, 0), (3, 0), (3, 3), (0, 3)]
    >>> lower_left_triangle = [(2, 1), (2, 2), (1, 2)]
    >>> lower_right_square = [(3, 0), (6, 0), (6, 3), (3, 3)]
    >>> lower_right_triangle = [(4, 1), (5, 2), (4, 2)]
    >>> upper_left_square = [(0, 3), (3, 3), (3, 6), (0, 6)]
    >>> upper_left_triangle = [(1, 4), (2, 4), (2, 5)]
    >>> upper_right_square = [(3, 3), (6, 3), (6, 6), (3, 6)]
    >>> upper_right_triangle = [(4, 4), (5, 4), (4, 5)]
    >>> unite_multipolygons([], [])
    []
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle])], [])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> unite_multipolygons([], [(lower_left_square, [lower_left_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 1), (2, 2), (1, 2)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                     [(lower_left_square, [lower_left_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                     [(lower_right_square, [lower_right_triangle])])
    [([(0, 0), (6, 0), (6, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)],\
 [(4, 2), (5, 2), (4, 1)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                     [(upper_left_square, [upper_left_triangle])])
    [([(0, 0), (3, 0), (3, 6), (0, 6)], [[(2, 2), (2, 1), (1, 2)],\
 [(2, 5), (2, 4), (1, 4)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle])],
    ...                     [(upper_right_square, [upper_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                      (upper_right_square, [upper_right_triangle])],
    ...                     [(upper_left_square, [upper_left_triangle]),
    ...                      (lower_right_square, [lower_right_triangle])])
    [([(0, 0), (6, 0), (6, 6), (0, 6)], [[(2, 2), (2, 1), (1, 2)],\
 [(2, 5), (2, 4), (1, 4)], [(4, 2), (5, 2), (4, 1)],\
 [(4, 5), (5, 4), (4, 4)]])]
    >>> unite_multipolygons([(lower_left_square, [lower_left_triangle]),
    ...                      (upper_right_square, [upper_right_triangle])],
    ...                     [(lower_left_square, [lower_left_triangle]),
    ...                      (upper_right_square, [upper_right_triangle])])
    [([(0, 0), (3, 0), (3, 3), (0, 3)], [[(2, 2), (2, 1), (1, 2)]]),\
 ([(3, 3), (6, 3), (6, 6), (3, 6)], [[(4, 5), (5, 4), (4, 4)]])]
    """
    return _holey.Union(left, right, accurate).compute()



from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from numbers import Rational
from operator import attrgetter
from typing import (Iterable,
                    List,
                    Union as Union_)

from reprit.base import generate_repr

from clipping.hints import (Mix,
                            Multipoint,
                            Multisegment,
                            Segment)
from . import bounding_box
from .event import BinaryEvent
from .events_queue import (LinearBinaryEventsQueue as BinaryEventsQueue,
                           NaryEventsQueue)
from .sweep_line import (BinarySweepLine,
                         NarySweepLine)
from .utils import (all_equal,
                    to_multisegment_base,
                    to_multisegment_x_max,
                    to_rational_multisegment)


def merge_segments(segments: List[Segment],
                   accurate: bool) -> Iterable[Segment]:
    if not segments:
        return
    if accurate and not issubclass(to_multisegment_base(segments), Rational):
        segments = to_rational_multisegment(segments)
    events_queue = NaryEventsQueue()
    events_queue.register_segments(segments)
    sweep_line = NarySweepLine()
    while events_queue:
        event = events_queue.pop()
        if event.is_right_endpoint:
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    events_queue.detect_intersection(below_event, above_event)
                yield event.segment
        elif event not in sweep_line:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            if above_event is not None:
                events_queue.detect_intersection(event, above_event)
            if below_event is not None:
                events_queue.detect_intersection(below_event, event)


class Operation(ABC):
    __slots__ = 'left', 'right', 'accurate', '_events_queue'

    def __init__(self,
                 left: Multisegment,
                 right: Multisegment,
                 accurate: bool) -> None:
        """
        Initializes operation.

        :param left: left operand.
        :param right: right operand.
        :param accurate:
            flag that tells whether to use slow but more accurate arithmetic
            for floating point numbers.
        """
        self.left = left
        self.right = right
        self.accurate = accurate
        self._events_queue = BinaryEventsQueue()

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union_[Multisegment, Mix]:
        """
        Computes result of the operation.
        """

    def fill_queue(self) -> None:
        self._events_queue.register_segments(self.left, True)
        self._events_queue.register_segments(self.right, False)

    def normalize_operands(self) -> None:
        left, right = self.left, self.right
        if (self.accurate
                and not issubclass(to_multisegment_base(left + right),
                                   Rational)):
            self.left, self.right = (to_rational_multisegment(left),
                                     to_rational_multisegment(right))

    def process_event(self, event: BinaryEvent,
                      sweep_line: BinarySweepLine) -> None:
        if event.is_right_endpoint:
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)
        else:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            if above_event is not None:
                self._events_queue.detect_intersection(event, above_event)
            if below_event is not None:
                self._events_queue.detect_intersection(below_event, event)

    def sweep(self) -> List[BinaryEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine()
        while events_queue:
            event = events_queue.pop()
            self.process_event(event, sweep_line)
            if event.is_right_endpoint:
                result.append(event.complement)
        return result


class Difference(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.left and self.right):
            return self.left
        left_bounding_box = bounding_box.from_multisegment(self.left)
        right_bounding_box = bounding_box.from_multisegment(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return self.left
        self.right = bounding_box.to_coupled_segments(left_bounding_box,
                                                      self.right)
        if not self.right:
            return self.left
        self.normalize_operands()
        return sorted(segment
                      for segment, events in groupby(self.sweep(),
                                                     key=event_to_segment)
                      if all(event.from_left for event in events))

    def sweep(self) -> List[BinaryEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine()
        left_x_max = to_multisegment_x_max(self.left)
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > left_x_max:
                break
            self.process_event(event, sweep_line)
            if event.is_right_endpoint:
                result.append(event.complement)
        return result


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.left and self.right):
            return []
        left_bounding_box = bounding_box.from_multisegment(self.left)
        right_bounding_box = bounding_box.from_multisegment(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return []
        self.left = bounding_box.to_coupled_segments(right_bounding_box,
                                                     self.left)
        self.right = bounding_box.to_coupled_segments(left_bounding_box,
                                                      self.right)
        if not (self.left and self.right):
            return []
        self.normalize_operands()
        return sorted(segment
                      for segment, events in groupby(self.sweep(),
                                                     key=event_to_segment)
                      if not all_equal(event.from_left for event in events))

    def sweep(self) -> List[BinaryEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine()
        min_max_x = min(to_multisegment_x_max(self.left),
                        to_multisegment_x_max(self.right))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, sweep_line)
            if event.is_right_endpoint:
                result.append(event.complement)
        return result


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> Mix:
        if not (self.left and self.right):
            return [], [], []
        left_bounding_box = bounding_box.from_multisegment(self.left)
        right_bounding_box = bounding_box.from_multisegment(self.right)
        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
            return [], [], []
        self.left = bounding_box.to_intersecting_segments(right_bounding_box,
                                                          self.left)
        self.right = bounding_box.to_intersecting_segments(left_bounding_box,
                                                           self.right)
        if not (self.left and self.right):
            return [], [], []
        self.normalize_operands()
        multipoint = []  # type: Multipoint
        multisegment = []  # type: Multisegment
        for start, same_start_events in groupby(sorted(self.sweep(),
                                                       key=event_to_segment),
                                                key=attrgetter('start')):
            same_start_events = list(same_start_events)
            if not all_equal(event.from_left for event in same_start_events):
                no_segment_found = True
                for event, next_event in zip(same_start_events,
                                             same_start_events[1:]):
                    if (event.from_left is not next_event.from_left
                            and event.segment == next_event.segment):
                        no_segment_found = False
                        if not event.is_right_endpoint:
                            multisegment.append(event.segment)
                if no_segment_found:
                    multipoint.append(start)
        return multipoint, multisegment, []

    def sweep(self) -> List[BinaryEvent]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = BinarySweepLine()
        min_max_x = min(to_multisegment_x_max(self.left),
                        to_multisegment_x_max(self.right))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, sweep_line)
            result.append(event)
        return result


class SymmetricDifference(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.left and self.right):
            return self.left or self.right
        elif bounding_box.disjoint_with(
                bounding_box.from_multisegment(self.left),
                bounding_box.from_multisegment(self.right)):
            result = self.left + self.right
            result.sort()
            return result
        self.normalize_operands()
        return sorted(segment
                      for segment, events in groupby(self.sweep(),
                                                     key=event_to_segment)
                      if all_equal(event.from_left for event in events))


class Union(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.left and self.right):
            return self.left or self.right
        elif bounding_box.disjoint_with(
                bounding_box.from_multisegment(self.left),
                bounding_box.from_multisegment(self.right)):
            result = self.left + self.right
            result.sort()
            return result
        self.normalize_operands()
        return sorted(segment
                      for segment, _ in groupby(self.sweep(),
                                                key=event_to_segment))


event_to_segment = attrgetter('segment')



from typing import (Iterable,
                    List,
                    Sequence)

from orient.planar import (Relation,
                           point_in_region,
                           segment_in_contour,
                           segment_in_region)
from robust.linear import (SegmentsRelationship,
                           segments_relationship)

from clipping.hints import (Contour,
                            Multipolygon,
                            Multiregion,
                            Multisegment,
                            Point,
                            Polygon,
                            Region,
                            Segment)
from .hints import BoundingBox
from .utils import (contour_to_segments,
                    flatten)


def from_points(points: Iterable[Point]) -> BoundingBox:
    """
    Builds bounding box from points.
    """
    points = iter(points)
    x_min, y_min = x_max, y_max = next(points)
    for x, y in points:
        if x < x_min:
            x_min = x
        elif x > x_max:
            x_max = x
        if y < y_min:
            y_min = y
        elif y > y_max:
            y_max = y
    return x_min, x_max, y_min, y_max


def from_multisegment(multisegment: Multisegment) -> BoundingBox:
    """
    Builds bounding box from multisegment.
    """
    return from_points(flatten(multisegment))


def from_multipolygon(multipolygon: Multipolygon) -> BoundingBox:
    """
    Builds bounding box from multipolygon.
    """
    return from_points(flatten(border for border, _ in multipolygon))


def from_multiregion(multiregion: Multiregion) -> BoundingBox:
    """
    Builds bounding box from multiregion.
    """
    return from_points(flatten(multiregion))


def disjoint_with(left: BoundingBox, right: BoundingBox) -> bool:
    """
    Checks if bounding boxes do not intersect.

    >>> disjoint_with((0, 2, 0, 2), (0, 2, 0, 2))
    False
    >>> disjoint_with((0, 2, 0, 2), (1, 3, 1, 3))
    False
    >>> disjoint_with((0, 2, 0, 2), (2, 4, 0, 2))
    False
    >>> disjoint_with((0, 2, 0, 2), (2, 4, 2, 4))
    False
    >>> disjoint_with((0, 2, 0, 2), (2, 4, 3, 5))
    True
    """
    left_x_min, left_x_max, left_y_min, left_y_max = left
    right_x_min, right_x_max, right_y_min, right_y_max = right
    return (left_x_min > right_x_max or left_x_max < right_x_min
            or left_y_min > right_y_max or left_y_max < right_y_min)


def intersects_with(left: BoundingBox, right: BoundingBox) -> bool:
    """
    Checks if bounding boxes intersect.

    >>> intersects_with((0, 2, 0, 2), (0, 2, 0, 2))
    True
    >>> intersects_with((0, 2, 0, 2), (1, 3, 1, 3))
    True
    >>> intersects_with((0, 2, 0, 2), (2, 4, 0, 2))
    True
    >>> intersects_with((0, 2, 0, 2), (2, 4, 2, 4))
    True
    >>> intersects_with((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    left_x_min, left_x_max, left_y_min, left_y_max = left
    right_x_min, right_x_max, right_y_min, right_y_max = right
    return (right_x_min <= left_x_max and left_x_min <= right_x_max
            and right_y_min <= left_y_max and left_y_min <= right_y_max)


def coupled_with(left: BoundingBox, right: BoundingBox) -> bool:
    """
    Checks if bounding boxes intersect in some region or by the edge.

    >>> coupled_with((0, 2, 0, 2), (0, 2, 0, 2))
    True
    >>> coupled_with((0, 2, 0, 2), (1, 3, 1, 3))
    True
    >>> coupled_with((0, 2, 0, 2), (2, 4, 0, 2))
    True
    >>> coupled_with((0, 2, 0, 2), (2, 4, 2, 4))
    False
    >>> coupled_with((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    return (intersects_with(left, right)
            and (not touches_with(left, right)
                 or edges_overlap_with(left, right)))


def touches_with(left: BoundingBox, right: BoundingBox) -> bool:
    """
    Checks if bounding boxes intersect at point or by the edge.

    >>> touches_with((0, 2, 0, 2), (0, 2, 0, 2))
    False
    >>> touches_with((0, 2, 0, 2), (1, 3, 1, 3))
    False
    >>> touches_with((0, 2, 0, 2), (2, 4, 0, 2))
    True
    >>> touches_with((0, 2, 0, 2), (2, 4, 2, 4))
    True
    >>> touches_with((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    left_x_min, left_x_max, left_y_min, left_y_max = left
    right_x_min, right_x_max, right_y_min, right_y_max = right
    return ((left_x_min == right_x_max or left_x_max == right_x_min)
            and (left_y_min <= right_y_max and right_y_min <= left_y_max)
            or (left_x_min <= right_x_max and right_x_min <= left_x_max)
            and (left_y_min == right_y_max or right_y_min == left_y_max))


def edges_overlap_with(left: BoundingBox, right: BoundingBox) -> bool:
    """
    Checks if bounding boxes intersect by the edge.

    >>> edges_overlap_with((0, 2, 0, 2), (0, 2, 0, 2))
    False
    >>> edges_overlap_with((0, 2, 0, 2), (1, 3, 1, 3))
    False
    >>> edges_overlap_with((0, 2, 0, 2), (2, 4, 0, 2))
    True
    >>> edges_overlap_with((0, 2, 0, 2), (2, 4, 2, 4))
    False
    >>> edges_overlap_with((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    left_x_min, left_x_max, left_y_min, left_y_max = left
    right_x_min, right_x_max, right_y_min, right_y_max = right
    return ((left_x_min == right_x_max or left_x_max == right_x_min)
            and (left_y_min < right_y_max and right_y_min < left_y_max)
            or (left_x_min < right_x_max and right_x_min < left_x_max)
            and (left_y_min == right_y_max or right_y_min == left_y_max))


def is_subset_of(test: BoundingBox, goal: BoundingBox) -> bool:
    """
    Checks if the bounding box is the subset of the other.

    >>> is_subset_of((0, 2, 0, 2), (0, 2, 0, 2))
    True
    >>> is_subset_of((0, 2, 0, 2), (1, 3, 1, 3))
    False
    >>> is_subset_of((0, 2, 0, 2), (2, 4, 0, 2))
    False
    >>> is_subset_of((0, 2, 0, 2), (2, 4, 2, 4))
    False
    >>> is_subset_of((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    test_x_min, test_x_max, test_y_min, test_y_max = test
    goal_x_min, goal_x_max, goal_y_min, goal_y_max = goal
    return (goal_x_min <= test_x_min and test_x_max <= goal_x_max
            and goal_y_min <= test_y_min and test_y_max <= goal_y_max)


def within_of(test: BoundingBox, goal: BoundingBox) -> bool:
    """
    Checks if the bounding box is contained in an interior of the other.

    >>> within_of((0, 2, 0, 2), (0, 2, 0, 2))
    False
    >>> within_of((0, 2, 0, 2), (1, 3, 1, 3))
    False
    >>> within_of((0, 2, 0, 2), (2, 4, 0, 2))
    False
    >>> within_of((0, 2, 0, 2), (2, 4, 2, 4))
    False
    >>> within_of((0, 2, 0, 2), (2, 4, 3, 5))
    False
    """
    test_x_min, test_x_max, test_y_min, test_y_max = test
    goal_x_min, goal_x_max, goal_y_min, goal_y_max = goal
    return (goal_x_min < test_x_min and test_x_max < goal_x_max
            and goal_y_min < test_y_min and test_y_max < goal_y_max)


def intersects_with_segment(bounding_box: BoundingBox,
                            segment: Segment) -> bool:
    """
    Checks if the bounding box intersects the segment.
    """
    segment_bounding_box = from_points(segment)
    return (intersects_with(segment_bounding_box, bounding_box)
            and (is_subset_of(segment_bounding_box, bounding_box)
                 or any(segments_relationship(edge, segment)
                        is not SegmentsRelationship.NONE
                        for edge in to_segments(bounding_box))))


def coupled_with_segment(bounding_box: BoundingBox,
                         segment: Segment) -> bool:
    """
    Checks if the bounding box intersects the segment at more than one point.
    """
    segment_bounding_box = from_points(segment)
    return (coupled_with(segment_bounding_box, bounding_box)
            and (is_subset_of(segment_bounding_box, bounding_box)
                 or any(segments_relationship(edge, segment)
                        not in (SegmentsRelationship.TOUCH,
                                SegmentsRelationship.NONE)
                        for edge in to_segments(bounding_box))))


def is_subset_of_region(bounding_box: BoundingBox, border: Contour) -> bool:
    """
    Checks if the bounding box is the subset of the region.
    """
    return all(segment_in_region(segment, border) in (Relation.COMPONENT,
                                                      Relation.ENCLOSED,
                                                      Relation.WITHIN)
               for segment in to_segments(bounding_box))


def within_of_region(bounding_box: BoundingBox, border: Contour) -> bool:
    """
    Checks if the bounding box is contained in an interior of the region.
    """
    return (all(point_in_region(vertex, border) is Relation.WITHIN
                for vertex in to_vertices(bounding_box))
            and all(segments_relationship(edge, border_edge)
                    is SegmentsRelationship.NONE
                    for edge in to_segments(bounding_box)
                    for border_edge in contour_to_segments(border)))


def is_subset_of_multiregion(bounding_box: BoundingBox,
                             borders: List[Contour]) -> bool:
    """
    Checks if the bounding box is the subset of the multiregion.
    """
    return any(is_subset_of(bounding_box, from_points(border))
               and is_subset_of_region(bounding_box, border)
               for border in borders)


def intersects_with_polygon(bounding_box: BoundingBox,
                            polygon: Polygon) -> bool:
    """
    Checks if the bounding box intersects the polygon.
    """
    border, holes = polygon
    polygon_bounding_box = from_points(border)
    if not intersects_with(polygon_bounding_box, bounding_box):
        return False
    elif (is_subset_of(polygon_bounding_box, bounding_box)
          or any(contains_point(bounding_box, vertex)
                 for vertex in border)):
        return True
    relations = [point_in_region(vertex, border)
                 for vertex in to_vertices(bounding_box)]
    if (within_of(bounding_box, polygon_bounding_box)
            and all(relation is Relation.WITHIN for relation in relations)
            and all(segments_relationship(edge, border_edge)
                    is SegmentsRelationship.NONE
                    for edge in to_segments(bounding_box)
                    for border_edge in contour_to_segments(border))):
        return not any(within_of(bounding_box, from_points(hole))
                       and within_of_region(bounding_box, hole)
                       for hole in holes)
    else:
        return (any(relation is not Relation.DISJOINT
                    for relation in relations)
                or any(intersects_with_segment(bounding_box, border_edge)
                       for border_edge in contour_to_segments(border)))


def intersects_with_region(bounding_box: BoundingBox,
                           region: Region) -> bool:
    """
    Checks if the bounding box intersects the region.
    """
    region_bounding_box = from_points(region)
    return (intersects_with(region_bounding_box, bounding_box)
            and (is_subset_of(region_bounding_box, bounding_box)
                 or any(contains_point(bounding_box, vertex)
                        for vertex in region)
                 or any(point_in_region(vertex, region)
                        is not Relation.DISJOINT
                        for vertex in to_vertices(bounding_box))
                 or any(intersects_with_segment(bounding_box, border_edge)
                        for border_edge in contour_to_segments(region))))


def coupled_with_polygon(bounding_box: BoundingBox, polygon: Polygon) -> bool:
    """
    Checks if the bounding box intersects the polygon in continuous points set.
    """
    border, holes = polygon
    polygon_bounding_box = from_points(border)
    if not coupled_with(polygon_bounding_box, bounding_box):
        return False
    elif (is_subset_of(polygon_bounding_box, bounding_box)
          or any(covers_point(bounding_box, vertex)
                 for vertex in border)):
        return True
    relations = [point_in_region(vertex, border)
                 for vertex in to_vertices(bounding_box)]
    if any(relation is Relation.WITHIN for relation in relations):
        return (not all(relation is Relation.WITHIN for relation in relations)
                or not is_subset_of_multiregion(bounding_box, holes))
    else:
        return (not is_subset_of_multiregion(bounding_box, holes)
                if (is_subset_of(bounding_box, polygon_bounding_box)
                    and is_subset_of_region(bounding_box, border))
                else any(segment_in_contour(segment, border)
                         is Relation.OVERLAP
                         or segment_in_region(segment, border)
                         in (Relation.CROSS, Relation.COMPONENT,
                             Relation.ENCLOSED)
                         for segment in to_segments(bounding_box)))


def coupled_with_region(bounding_box: BoundingBox, region: Region) -> bool:
    """
    Checks if the bounding box intersects the region in continuous points set.
    """
    region_bounding_box = from_points(region)
    if not coupled_with(region_bounding_box, bounding_box):
        return False
    elif (is_subset_of(region_bounding_box, bounding_box)
          or any(covers_point(bounding_box, vertex)
                 for vertex in region)):
        return True
    return (any(point_in_region(vertex, region) is Relation.WITHIN
                for vertex in to_vertices(bounding_box))
            or is_subset_of(bounding_box, region_bounding_box)
            and is_subset_of_region(bounding_box, region)
            or any(segment_in_contour(segment, region)
                   is Relation.OVERLAP
                   or segment_in_region(segment, region)
                   in (Relation.CROSS, Relation.COMPONENT,
                       Relation.ENCLOSED)
                   for segment in to_segments(bounding_box)))


def contains_point(bounding_box: BoundingBox, point: Point) -> bool:
    x_min, x_max, y_min, y_max = bounding_box
    x, y = point
    return x_min <= x <= x_max and y_min <= y <= y_max


def covers_point(bounding_box: BoundingBox, point: Point) -> bool:
    x_min, x_max, y_min, y_max = bounding_box
    x, y = point
    return x_min < x < x_max and y_min < y < y_max


def to_vertices(bounding_box: BoundingBox) -> Sequence[Point]:
    x_min, x_max, y_min, y_max = bounding_box
    return (x_min, y_min), (x_max, y_min), (x_max, y_max), (x_min, y_max)


def to_segments(bounding_box: BoundingBox) -> Sequence[Segment]:
    x_min, x_max, y_min, y_max = bounding_box
    return (((x_min, y_min), (x_max, y_min)),
            ((x_max, y_min), (x_max, y_max)),
            ((x_min, y_max), (x_max, y_max)),
            ((x_min, y_min), (x_min, y_max)))


def to_intersecting_segments(bounding_box: BoundingBox,
                             multisegment: Multisegment) -> Multisegment:
    return [segment
            for segment in multisegment
            if intersects_with_segment(bounding_box, segment)]


def to_coupled_segments(bounding_box: BoundingBox,
                        multisegment: Multisegment) -> Multisegment:
    return [segment
            for segment in multisegment
            if coupled_with_segment(bounding_box, segment)]


def to_intersecting_polygons(bounding_box: BoundingBox,
                             multipolygon: Multipolygon) -> Multipolygon:
    return [polygon
            for polygon in multipolygon
            if intersects_with_polygon(bounding_box, polygon)]


def to_intersecting_regions(bounding_box: BoundingBox,
                            multiregion: Multiregion) -> Multiregion:
    return [region
            for region in multiregion
            if intersects_with_region(bounding_box, region)]


def to_coupled_polygons(bounding_box: BoundingBox,
                        multipolygon: Multipolygon) -> Multipolygon:
    return [polygon
            for polygon in multipolygon
            if coupled_with_polygon(bounding_box, polygon)]


def to_coupled_regions(bounding_box: BoundingBox,
                       multiregion: Multiregion) -> Multiregion:
    return [region
            for region in multiregion
            if coupled_with_region(bounding_box, region)]



from abc import (ABC,
                 abstractmethod)
from itertools import groupby
from numbers import Rational
from operator import attrgetter
from typing import (List,
                    Optional,
                    Union)

from reprit.base import generate_repr

from clipping.hints import (Mix,
                            Multipoint,
                            Multipolygon,
                            Multisegment)
from . import bounding_box
from .event import MixedEvent as Event
from .events_queue import (BinaryEventsQueueKey as EventsQueueKey,
                           MixedBinaryEventsQueue as EventsQueue)
from .sweep_line import BinarySweepLine as SweepLine
from .utils import (all_equal,
                    polygon_to_oriented_segments,
                    to_mixed_base,
                    to_multipolygon_x_max,
                    to_multisegment_x_max,
                    to_rational_multipolygon,
                    to_rational_multisegment)


class Operation(ABC):
    __slots__ = 'multisegment', 'multipolygon', 'accurate', '_events_queue'

    def __init__(self,
                 multisegment: Multisegment,
                 multipolygon: Multipolygon,
                 accurate: bool) -> None:
        """
        Initializes operation.

        :param multisegment: left operand.
        :param multipolygon: right operand.
        :param accurate:
            flag that tells whether to use slow but more accurate arithmetic
            for floating point numbers.
        """
        self.multisegment = multisegment
        self.multipolygon = multipolygon
        self.accurate = accurate
        self._events_queue = EventsQueue()

    __repr__ = generate_repr(__init__)

    @abstractmethod
    def compute(self) -> Union[Multisegment, Mix]:
        """
        Computes result of the operation.
        """

    def compute_fields(self, event: Event,
                       below_event: Optional[Event]) -> None:
        if below_event is not None:
            event.other_interior_to_left = (below_event.other_interior_to_left
                                            if (event.from_left
                                                is below_event.from_left)
                                            else below_event.interior_to_left)
        event.in_result = self.in_result(event)

    def fill_queue(self) -> None:
        events_queue = self._events_queue
        events_queue.register_segments(self.multisegment, True)
        for polygon in self.multipolygon:
            events_queue.register_segments(
                    polygon_to_oriented_segments(polygon), False)

    @abstractmethod
    def in_result(self, event: Event) -> bool:
        """Detects if event will be presented in result of the operation."""

    def normalize_operands(self) -> None:
        multisegment, multipolygon = self.multisegment, self.multipolygon
        if (self.accurate
                and not issubclass(to_mixed_base(multisegment, multipolygon),
                                   Rational)):
            self.multisegment, self.multipolygon = (
                to_rational_multisegment(multisegment),
                to_rational_multipolygon(multipolygon))

    def process_event(self, event: Event, sweep_line: SweepLine) -> None:
        if event.is_right_endpoint:
            event = event.complement
            if event in sweep_line:
                above_event, below_event = (sweep_line.above(event),
                                            sweep_line.below(event))
                sweep_line.remove(event)
                if above_event is not None and below_event is not None:
                    self._events_queue.detect_intersection(below_event,
                                                           above_event)
        elif event not in sweep_line:
            sweep_line.add(event)
            above_event, below_event = (sweep_line.above(event),
                                        sweep_line.below(event))
            self.compute_fields(event, below_event)
            if (above_event is not None
                    and self._events_queue.detect_intersection(event,
                                                               above_event)):
                self.compute_fields(event, below_event)
                self.compute_fields(above_event, event)
            if (below_event is not None
                    and self._events_queue.detect_intersection(below_event,
                                                               event)):
                below_below_event = sweep_line.below(below_event)
                self.compute_fields(below_event, below_below_event)
                self.compute_fields(event, below_event)


class Difference(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.multisegment and self.multipolygon):
            return self.multisegment
        multisegment_bounding_box = (bounding_box
                                     .from_multisegment(self.multisegment))
        if bounding_box.disjoint_with(
                multisegment_bounding_box,
                bounding_box.from_multipolygon(self.multipolygon)):
            return self.multisegment
        self.multipolygon = bounding_box.to_coupled_polygons(
                multisegment_bounding_box, self.multipolygon)
        if not self.multipolygon:
            return self.multisegment
        self.normalize_operands()
        return [event.segment for event in self.sweep() if event.in_result]

    def in_result(self, event: Event) -> bool:
        return event.from_left and event.outside

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        left_x_max = to_multisegment_x_max(self.multisegment)
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > left_x_max:
                break
            self.process_event(event, sweep_line)
            if event.is_right_endpoint:
                result.append(event.complement)
        return result


class CompleteIntersection(Operation):
    __slots__ = ()

    def compute(self) -> Mix:
        if not (self.multisegment and self.multipolygon):
            return [], [], []
        multisegment_bounding_box = (bounding_box
                                     .from_multisegment(self.multisegment))
        multipolygon_bounding_box = (bounding_box
                                     .from_multipolygon(self.multipolygon))
        if bounding_box.disjoint_with(multisegment_bounding_box,
                                      multipolygon_bounding_box):
            return [], [], []
        self.multisegment = bounding_box.to_intersecting_segments(
                multipolygon_bounding_box, self.multisegment)
        self.multipolygon = bounding_box.to_intersecting_polygons(
                multisegment_bounding_box, self.multipolygon)
        if not (self.multisegment and self.multipolygon):
            return [], [], []
        self.normalize_operands()
        events = sorted(self.sweep(),
                        key=EventsQueueKey)
        multipoint = []  # type: Multipoint
        border_multisegment = []  # type: Multisegment
        for start, same_start_events in groupby(events,
                                                key=attrgetter('start')):
            same_start_events = list(same_start_events)
            if (all(not (event.in_result or event.is_right_endpoint
                         and event.complement.in_result)
                    for event in same_start_events)
                    and not all_equal(event.from_left
                                      for event in same_start_events)):
                multipoint.append(start)
        inside_multisegment = [event.segment
                               for event in events
                               if event.in_result]
        return multipoint, border_multisegment + inside_multisegment, []

    def in_result(self, event: Event) -> bool:
        return event.from_left and not event.outside

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        min_max_x = min(to_multisegment_x_max(self.multisegment),
                        to_multipolygon_x_max(self.multipolygon))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, sweep_line)
            result.append(event)
        return result


class Intersection(Operation):
    __slots__ = ()

    def compute(self) -> Multisegment:
        if not (self.multisegment and self.multipolygon):
            return []
        multisegment_bounding_box = (bounding_box
                                     .from_multisegment(self.multisegment))
        multipolygon_bounding_box = (bounding_box
                                     .from_multipolygon(self.multipolygon))
        if bounding_box.disjoint_with(multisegment_bounding_box,
                                      multipolygon_bounding_box):
            return []
        self.multisegment = bounding_box.to_intersecting_segments(
                multipolygon_bounding_box, self.multisegment)
        self.multipolygon = bounding_box.to_intersecting_polygons(
                multisegment_bounding_box, self.multipolygon)
        if not (self.multisegment and self.multipolygon):
            return []
        self.normalize_operands()
        return [event.segment for event in self.sweep() if event.in_result]

    def in_result(self, event: Event) -> bool:
        return event.from_left and not event.outside

    def sweep(self) -> List[Event]:
        self.fill_queue()
        result = []
        events_queue = self._events_queue
        sweep_line = SweepLine()
        min_max_x = min(to_multisegment_x_max(self.multisegment),
                        to_multipolygon_x_max(self.multipolygon))
        while events_queue:
            event = events_queue.pop()
            start_x, _ = event.start
            if start_x > min_max_x:
                break
            self.process_event(event, sweep_line)
            if not event.is_right_endpoint:
                result.append(event)
        return result



from numbers import Real
from typing import (List,
                    Tuple,
                    Type)

Coordinate = Real
Base = Type[Coordinate]
Point = Tuple[Coordinate, Coordinate]
Multipoint = List[Point]
Segment = Tuple[Point, Point]
Multisegment = List[Segment]
Contour = Region = List[Point]
Multiregion = List[Region]
Polygon = Tuple[Region, Multiregion]
Multipolygon = List[Polygon]
Mix = Tuple[Multipoint, Multisegment, Multipolygon]
HolelessMix = Tuple[Multipoint, Multisegment, Multiregion]



from fractions import Fraction
from itertools import (chain,
                       groupby)
from numbers import (Integral,
                     Number,
                     Rational,
                     Real)
from typing import (Any,
                    Iterable,
                    Sequence,
                    Tuple,
                    Type,
                    TypeVar)

from robust.angular import (Orientation,
                            orientation)

from clipping.hints import (Base,
                            Contour,
                            Coordinate,
                            Multipolygon,
                            Multiregion,
                            Multisegment,
                            Point,
                            Polygon,
                            Segment)


def all_equal(iterable: Iterable[Any]) -> bool:
    groups = groupby(iterable)
    return next(groups, True) and not next(groups, False)


Domain = TypeVar('Domain')


def pairwise(iterable: Iterable[Domain]) -> Iterable[Tuple[Domain, Domain]]:
    iterator = iter(iterable)
    element = next(iterator, None)
    for next_element in iterator:
        yield element, next_element
        element = next_element


def to_multipolygon_contours(multipolygon: Multipolygon) -> Iterable[Contour]:
    for border, holes in multipolygon:
        yield border
        yield from holes


def polygon_to_oriented_segments(polygon: Polygon) -> Iterable[Segment]:
    border, holes = polygon
    yield from contour_to_oriented_segments(border,
                                            clockwise=False)
    for hole in holes:
        yield from contour_to_oriented_segments(hole,
                                                clockwise=True)


def contour_to_oriented_segments(contour: Contour,
                                 *,
                                 clockwise: bool = False) -> Iterable[Segment]:
    return (((contour[index - 1], contour[index])
             for index in range(len(contour)))
            if (to_contour_orientation(contour)
                is (Orientation.CLOCKWISE
                    if clockwise
                    else Orientation.COUNTERCLOCKWISE))
            else ((contour[index], contour[index - 1])
                  for index in range(len(contour) - 1, -1, -1)))


def contour_to_segments(contour: Contour) -> Iterable[Segment]:
    return ((contour[index - 1], contour[index])
            for index in range(len(contour)))


def to_contour_orientation(contour: Contour) -> Orientation:
    index = min(range(len(contour)),
                key=contour.__getitem__)
    return orientation(contour[index], contour[index - 1],
                       contour[(index + 1) % len(contour)])


def to_mixed_base(multisegment: Multisegment,
                  multipolygon: Multipolygon) -> Base:
    return max(to_multisegment_base(multisegment),
               to_multipolygon_base(multipolygon),
               key=_bases_key)


def to_multipolygon_base(multipolygon: Multipolygon) -> Base:
    return max({to_contour_base(contour)
                for contour in to_multipolygon_contours(multipolygon)},
               key=_bases_key)


def to_multiregion_base(multipolygon: Multiregion) -> Base:
    return max({to_contour_base(region) for region in multipolygon},
               key=_bases_key)


def to_multisegment_base(multisegment: Multisegment) -> Base:
    return max(set(map(type, flatten(flatten(multisegment)))),
               key=_bases_key)


def to_contour_base(contour: Contour) -> Base:
    return max(set(map(type, flatten(contour))),
               key=_bases_key)


flatten = chain.from_iterable


def _bases_key(base: Base,
               types: Sequence[Type[Number]] = (Integral, Rational, Real)
               ) -> int:
    try:
        return next(index
                    for index, type_ in enumerate(types)
                    if issubclass(base, type_))
    except StopIteration:
        if issubclass(base, Number):
            raise TypeError('{type} numbers are not supported.'
                            .format(type=base))
        else:
            raise TypeError('{type} is not recognized as a number type.'
                            .format(type=base))


def to_rational_multipolygon(multipolygon: Multipolygon) -> Multipolygon:
    return [(to_rational_contour(border), [to_rational_contour(hole)
                                           for hole in holes])
            for border, holes in multipolygon]


def to_rational_multiregion(multiregion: Multiregion) -> Multiregion:
    return [to_rational_contour(region) for region in multiregion]


def to_rational_multisegment(multisegment: Multisegment) -> Multisegment:
    return [(to_rational_point(start), to_rational_point(end))
            for start, end in multisegment]


def to_rational_contour(contour: Contour) -> Contour:
    return [to_rational_point(vertex) for vertex in contour]


def to_rational_point(point: Point) -> Point:
    x, y = point
    return Fraction(x), Fraction(y)


def to_first_boundary_vertex(polygon: Polygon) -> Point:
    boundary, _ = polygon
    return boundary[0]


def to_multipolygon_x_max(multipolygon: Multipolygon) -> Coordinate:
    return max(x for border, _ in multipolygon for x, _ in border)


def to_multiregion_x_max(multiregion: Multiregion) -> Coordinate:
    return max(x for border in multiregion for x, _ in border)


def to_multisegment_x_max(multisegment: Multisegment) -> Coordinate:
    return max(x for x, _ in flatten(multisegment))


def shrink_collinear_vertices(contour: Contour) -> None:
    index = -len(contour) + 1
    while index < 0:
        while (max(2, -index) < len(contour)
               and (orientation(contour[index + 2], contour[index + 1],
                                contour[index])
                    is Orientation.COLLINEAR)):
            del contour[index + 1]
        index += 1
    while index < len(contour):
        while (max(2, index) < len(contour)
               and (orientation(contour[index - 2], contour[index - 1],
                                contour[index])
                    is Orientation.COLLINEAR)):
            del contour[index - 1]
        index += 1



from functools import partial
from typing import (Callable,
                    Generic,
                    Iterable,
                    Type,
                    cast)

from prioq.base import PriorityQueue
from reprit.base import generate_repr
from robust.angular import (Orientation,
                            orientation)
from robust.linear import (SegmentsRelationship,
                           segments_intersection,
                           segments_relationship)

from clipping.hints import (Point,
                            Segment)
from .enums import OverlapKind
from .event import (BinaryEvent,
                    Event,
                    HoleyEvent,
                    MixedEvent,
                    NaryEvent,
                    ShapedEvent)


class BinaryEventsQueueKey:
    __slots__ = 'event',

    def __init__(self, event: BinaryEvent) -> None:
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'BinaryEventsQueueKey') -> bool:
        event, other_event = self.event, other.event
        start_x, start_y = event.start
        other_start_x, other_start_y = other_event.start
        if start_x != other_start_x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start_x < other_start_x
        elif start_y != other_start_y:
            # different points, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start_y < other_start_y
        elif event.is_right_endpoint is not other_event.is_right_endpoint:
            # same start, but one is a left endpoint
            # and the other a right endpoint,
            # the right endpoint is processed first
            return event.is_right_endpoint
        # same start, both events are left endpoints
        # or both are right endpoints
        else:
            other_end_orientation = orientation(event.end, event.start,
                                                other_event.end)
            # the lowest segment is processed first
            return (other_event.from_left
                    if other_end_orientation is Orientation.COLLINEAR
                    else (other_end_orientation
                          # the lowest segment is processed first
                          is (Orientation.CLOCKWISE
                              if event.is_right_endpoint
                              else Orientation.COUNTERCLOCKWISE)))


class NaryEventsQueueKey:
    __slots__ = 'event',

    def __init__(self, event: NaryEvent) -> None:
        self.event = event

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: 'NaryEventsQueueKey') -> bool:
        event, other_event = self.event, other.event
        start_x, start_y = event.start
        other_start_x, other_start_y = other_event.start
        if start_x != other_start_x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start_x < other_start_x
        elif start_y != other_start_y:
            # different points, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start_y < other_start_y
        elif event.is_right_endpoint is not other_event.is_right_endpoint:
            # same start, but one is a left endpoint
            # and the other a right endpoint,
            # the right endpoint is processed first
            return event.is_right_endpoint
        # same start, both events are left endpoints
        # or both are right endpoints
        else:
            # the lowest segment is processed first
            return (orientation(event.start, other_event.end, event.end)
                    is (Orientation.CLOCKWISE
                        if event.is_right_endpoint
                        else Orientation.COUNTERCLOCKWISE))


class LinearBinaryEventsQueue:
    __slots__ = '_queue',

    def __init__(self) -> None:
        self._queue = PriorityQueue(key=BinaryEventsQueueKey)

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self,
                            below_event: BinaryEvent,
                            event: BinaryEvent) -> None:
        below_segment, segment = below_event.segment, event.segment
        relationship = segments_relationship(below_segment, segment)
        if relationship is SegmentsRelationship.OVERLAP:
            # segments overlap
            if below_event.from_left is event.from_left:
                raise ValueError('Segments of the same multisegment '
                                 'should not overlap.')
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif (BinaryEventsQueueKey(event)
                  < BinaryEventsQueueKey(below_event)):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif (BinaryEventsQueueKey(event.complement)
                  < BinaryEventsQueueKey(below_event.complement)):
                end_min, end_max = event.complement, below_event.complement
            else:
                end_min, end_max = below_event.complement, event.complement
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.complement, end_min.start)
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.complement:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        elif (relationship is not SegmentsRelationship.NONE
              and event.start != below_event.start
              and event.end != below_event.end):
            # segments do not intersect_multipolygons at endpoints
            point = segments_intersection(below_segment, segment)
            if point != below_event.start and point != below_event.end:
                self._divide_segment(below_event, point)
            if point != event.start and point != event.end:
                self._divide_segment(event, point)

    def pop(self) -> BinaryEvent:
        return self._queue.pop()

    def register_segments(self,
                          segments: Iterable[Segment],
                          from_left: bool) -> None:
        events_queue = self._queue
        for start, end in segments:
            if start > end:
                start, end = end, start
            start_event = BinaryEvent(False, start, None, from_left)
            end_event = BinaryEvent(True, end, start_event, from_left)
            start_event.complement = end_event
            events_queue.push(start_event)
            events_queue.push(end_event)

    def _divide_segment(self, event: BinaryEvent, point: Point) -> None:
        left_event = BinaryEvent(False, point, event.complement,
                                 event.from_left)
        right_event = BinaryEvent(True, point, event, event.from_left)
        event.complement.complement, event.complement = left_event, right_event
        self._queue.push(left_event)
        self._queue.push(right_event)


class MixedBinaryEventsQueue:
    __slots__ = '_queue'

    def __init__(self) -> None:
        self._queue = PriorityQueue(key=BinaryEventsQueueKey)

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self,
                            below_event: MixedEvent,
                            event: MixedEvent) -> bool:
        below_segment, segment = below_event.segment, event.segment
        relationship = segments_relationship(below_segment, segment)
        if relationship is SegmentsRelationship.OVERLAP:
            # segments overlap
            if below_event.from_left is event.from_left:
                raise ValueError('Edges of the {geometry} '
                                 'should not overlap.'
                                 .format(geometry=('multisegment'
                                                   if event.from_left
                                                   else 'multipolygon')))
            event.is_overlap = below_event.is_overlap = True
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif (BinaryEventsQueueKey(event)
                  < BinaryEventsQueueKey(below_event)):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif (BinaryEventsQueueKey(event.complement)
                  < BinaryEventsQueueKey(below_event.complement)):
                end_min, end_max = event.complement, below_event.complement
            else:
                end_min, end_max = below_event.complement, event.complement
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.complement, end_min.start)
                return True
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.complement:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        elif (relationship is not SegmentsRelationship.NONE
              and event.start != below_event.start
              and event.end != below_event.end):
            # segments do not intersect_multipolygons at endpoints
            point = segments_intersection(below_segment, segment)
            if point != below_event.start and point != below_event.end:
                self._divide_segment(below_event, point)
            if point != event.start and point != event.end:
                self._divide_segment(event, point)
        return False

    def pop(self) -> MixedEvent:
        return self._queue.pop()

    def register_segments(self,
                          segments: Iterable[Segment],
                          from_left: bool) -> None:
        queue = self._queue
        for start, end in segments:
            interior_to_left = True
            if start > end:
                start, end = end, start
                interior_to_left = False
            start_event = MixedEvent(False, start, None, from_left,
                                     interior_to_left)
            end_event = MixedEvent(True, end, start_event, from_left,
                                   interior_to_left)
            start_event.complement = end_event
            queue.push(start_event)
            queue.push(end_event)

    def _divide_segment(self, event: MixedEvent, point: Point) -> None:
        left_event = MixedEvent(False, point, event.complement,
                                event.from_left, event.interior_to_left)
        right_event = MixedEvent(True, point, event, event.from_left,
                                 event.interior_to_left)
        event.complement.complement, event.complement = left_event, right_event
        self._queue.push(left_event)
        self._queue.push(right_event)


class NaryEventsQueue:
    __slots__ = '_queue',

    def __init__(self) -> None:
        self._queue = PriorityQueue(key=NaryEventsQueueKey)

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self,
                            below_event: NaryEvent,
                            event: NaryEvent) -> None:
        below_segment, segment = below_event.segment, event.segment
        relationship = segments_relationship(below_segment, segment)
        if relationship is SegmentsRelationship.OVERLAP:
            # segments overlap
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif NaryEventsQueueKey(event) < NaryEventsQueueKey(below_event):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif (NaryEventsQueueKey(event.complement)
                  < NaryEventsQueueKey(below_event.complement)):
                end_min, end_max = event.complement, below_event.complement
            else:
                end_min, end_max = below_event.complement, event.complement
            if starts_equal:
                # both line segments are equal or share the left endpoint
                if not ends_equal:
                    self._divide_segment(end_max.complement, end_min.start)
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.complement:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        elif (relationship is not SegmentsRelationship.NONE
              and event.start != below_event.start
              and event.end != below_event.end):
            # segments do not intersect_multipolygons at endpoints
            point = segments_intersection(below_segment, segment)
            if point != below_event.start and point != below_event.end:
                self._divide_segment(below_event, point)
            if point != event.start and point != event.end:
                self._divide_segment(event, point)

    def pop(self) -> NaryEvent:
        return self._queue.pop()

    def register_segments(self, segments: Iterable[Segment]) -> None:
        queue = self._queue
        for start, end in segments:
            if start > end:
                start, end = end, start
            start_event = NaryEvent(False, start, None)
            end_event = NaryEvent(True, end, start_event)
            start_event.complement = end_event
            queue.push(start_event)
            queue.push(end_event)

    def _divide_segment(self, event: NaryEvent, point: Point) -> None:
        left_event = NaryEvent(False, point, event.complement)
        right_event = NaryEvent(True, point, event)
        event.complement.complement, event.complement = left_event, right_event
        self._queue.push(left_event)
        self._queue.push(right_event)


class ShapedBinaryEventsQueue(Generic[Event]):
    __slots__ = 'event_cls', '_queue'

    def __init__(self, event_cls: Type[Event]) -> None:
        self.event_cls = event_cls
        self._queue = PriorityQueue(key=BinaryEventsQueueKey)

    __repr__ = generate_repr(__init__)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def detect_intersection(self, below_event: Event, event: Event) -> bool:
        below_segment, segment = below_event.segment, event.segment
        relationship = segments_relationship(below_segment, segment)
        if relationship is SegmentsRelationship.OVERLAP:
            # segments overlap
            if below_event.from_left is event.from_left:
                raise ValueError('Edges of the same multipolygon '
                                 'should not overlap.')
            starts_equal = below_event.start == event.start
            if starts_equal:
                start_min = start_max = None
            elif (BinaryEventsQueueKey(event)
                  < BinaryEventsQueueKey(below_event)):
                start_min, start_max = event, below_event
            else:
                start_min, start_max = below_event, event
            ends_equal = event.end == below_event.end
            if ends_equal:
                end_min = end_max = None
            elif (BinaryEventsQueueKey(event.complement)
                  < BinaryEventsQueueKey(below_event.complement)):
                end_min, end_max = event.complement, below_event.complement
            else:
                end_min, end_max = below_event.complement, event.complement
            if starts_equal:
                # both line segments are equal or share the left endpoint
                below_event.overlap_kind = event.overlap_kind = (
                    OverlapKind.SAME_ORIENTATION
                    if event.interior_to_left is below_event.interior_to_left
                    else OverlapKind.DIFFERENT_ORIENTATION)
                if not ends_equal:
                    self._divide_segment(end_max.complement, end_min.start)
                return True
            elif ends_equal:
                # the line segments share the right endpoint
                self._divide_segment(start_min, start_max.start)
            elif start_min is end_max.complement:
                # one line segment includes the other one
                self._divide_segment(start_min, end_min.start)
                self._divide_segment(start_min, start_max.start)
            else:
                # no line segment includes the other one
                self._divide_segment(start_max, end_min.start)
                self._divide_segment(start_min, start_max.start)
        elif (relationship is not SegmentsRelationship.NONE
              and event.start != below_event.start
              and event.end != below_event.end):
            # segments do not intersect_multipolygons at endpoints
            point = segments_intersection(below_segment, segment)
            if point != below_event.start and point != below_event.end:
                self._divide_segment(below_event, point)
            if point != event.start and point != event.end:
                self._divide_segment(event, point)
        return False

    def pop(self) -> Event:
        return self._queue.pop()

    def register_segments(self,
                          segments: Iterable[Segment],
                          from_left: bool) -> None:
        event_cls = self.event_cls
        queue = self._queue
        for start, end in segments:
            inside_on_left = True
            if start > end:
                start, end = end, start
                inside_on_left = False
            start_event = event_cls(False, start, None, from_left,
                                    inside_on_left)
            end_event = event_cls(True, end, start_event, from_left,
                                  inside_on_left)
            start_event.complement = end_event
            queue.push(start_event)
            queue.push(end_event)

    def _divide_segment(self, event: Event, point: Point) -> None:
        left_event = self.event_cls(False, point, event.complement,
                                    event.from_left, event.interior_to_left)
        right_event = self.event_cls(True, point, event, event.from_left,
                                     event.interior_to_left)
        event.complement.complement, event.complement = left_event, right_event
        self._queue.push(left_event)
        self._queue.push(right_event)


HolelessEventsQueue = cast(Callable[[], ShapedBinaryEventsQueue[ShapedEvent]],
                           partial(ShapedBinaryEventsQueue, ShapedEvent))
HoleyEventsQueue = cast(Callable[[], ShapedBinaryEventsQueue[HoleyEvent]],
                        partial(ShapedBinaryEventsQueue, HoleyEvent))



from reprlib import recursive_repr
from typing import (List,
                    Optional,
                    TypeVar)

from reprit.base import generate_repr

from clipping.hints import (Point,
                            Segment)
from .enums import OverlapKind


class NaryEvent:
    __slots__ = 'is_right_endpoint', 'start', 'complement'

    def __init__(self,
                 is_right_endpoint: bool,
                 start: Point,
                 complement: Optional['Event']) -> None:
        self.is_right_endpoint = is_right_endpoint
        self.start = start
        self.complement = complement

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def end(self) -> Point:
        return self.complement.start

    @property
    def segment(self) -> Segment:
        return self.start, self.end


class BinaryEvent:
    __slots__ = 'is_right_endpoint', 'start', 'complement', 'from_left'

    def __init__(self,
                 is_right_endpoint: bool,
                 start: Point,
                 complement: Optional['Event'],
                 from_left: bool) -> None:
        self.is_right_endpoint = is_right_endpoint
        self.start = start
        self.complement = complement
        self.from_left = from_left

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def end(self) -> Point:
        return self.complement.start

    @property
    def is_vertical(self) -> bool:
        start_x, _ = self.start
        end_x, _ = self.end
        return start_x == end_x

    @property
    def primary(self) -> Optional['BinaryEvent']:
        return self.complement if self.is_right_endpoint else self

    @property
    def segment(self) -> Segment:
        return self.start, self.end


class MixedEvent(BinaryEvent):
    __slots__ = ('interior_to_left', 'other_interior_to_left', 'is_overlap',
                 'in_result')

    def __init__(self,
                 is_right_endpoint: bool,
                 start: Point,
                 complement: Optional['MixedEvent'],
                 from_left: bool,
                 interior_to_left: bool,
                 other_interior_to_left: bool = False,
                 is_overlap: bool = False,
                 in_result: bool = False) -> None:
        super().__init__(is_right_endpoint, start, complement, from_left)
        self.interior_to_left = interior_to_left
        self.other_interior_to_left = other_interior_to_left
        self.is_overlap = is_overlap
        self.in_result = in_result

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def outside(self) -> bool:
        """
        Checks if the segment touches or disjoint with the intersection.
        """
        return not self.other_interior_to_left and not self.is_overlap


class ShapedEvent(BinaryEvent):
    __slots__ = ('interior_to_left', 'other_interior_to_left', 'overlap_kind',
                 'in_result', 'position')

    def __init__(self,
                 is_right_endpoint: bool,
                 start: Point,
                 complement: Optional['ShapedEvent'],
                 from_left: bool,
                 interior_to_left: bool,
                 other_interior_to_left: bool = False,
                 overlap_kind: OverlapKind = OverlapKind.NONE,
                 in_result: bool = False,
                 position: int = 0) -> None:
        super().__init__(is_right_endpoint, start, complement, from_left)
        self.interior_to_left = interior_to_left
        self.other_interior_to_left = other_interior_to_left
        self.overlap_kind = overlap_kind
        self.in_result = in_result
        self.position = position

    __repr__ = recursive_repr()(generate_repr(__init__))

    @property
    def inside(self) -> bool:
        """
        Checks if the segment enclosed by
        or lies within the region of the intersection.
        """
        return (self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)

    @property
    def is_common_polyline_component(self) -> bool:
        """
        Checks if the segment is a component of intersection's polyline.
        """
        return self.overlap_kind is OverlapKind.DIFFERENT_ORIENTATION

    @property
    def is_common_region_boundary(self) -> bool:
        """
        Checks if the segment is a boundary of intersection's region.
        """
        return self.overlap_kind is OverlapKind.SAME_ORIENTATION

    @property
    def is_overlap(self) -> bool:
        """
        Checks if the segment lies on the boundary of both operands.
        """
        return self.overlap_kind is not OverlapKind.NONE

    @property
    def outside(self) -> bool:
        """
        Checks if the segment touches or disjoint with the intersection.
        """
        return (not self.other_interior_to_left
                and self.overlap_kind is OverlapKind.NONE)


class HoleyEvent(ShapedEvent):
    __slots__ = ('interior_to_left', 'other_interior_to_left', 'overlap_kind',
                 'in_result', 'result_in_out', 'position', 'contour_id',
                 'below_in_result_event')

    def __init__(self,
                 is_right_endpoint: bool,
                 start: Point,
                 complement: Optional['HoleyEvent'],
                 from_left: bool,
                 interior_to_left: bool,
                 other_interior_to_left: bool = False,
                 overlap_kind: OverlapKind = OverlapKind.NONE,
                 in_result: bool = False,
                 result_in_out: bool = False,
                 position: int = 0,
                 contour_id: Optional[int] = None,
                 below_in_result_event: Optional['HoleyEvent'] = None) -> None:
        super().__init__(is_right_endpoint, start, complement, from_left,
                         interior_to_left, other_interior_to_left,
                         overlap_kind, in_result, position)
        self.result_in_out = result_in_out
        self.contour_id = contour_id
        self.below_in_result_event = below_in_result_event

    __repr__ = recursive_repr()(generate_repr(__init__))


Event = TypeVar('Event', NaryEvent, BinaryEvent, MixedEvent, ShapedEvent,
                HoleyEvent)


def events_to_connectivity(events: List[BinaryEvent]) -> List[int]:
    events_count = len(events)
    result = [0] * events_count
    index = 0
    while index < events_count:
        current_start = events[index].start
        right_start_index = index
        while (index < events_count
               and events[index].start == current_start
               and events[index].is_right_endpoint):
            index += 1
        right_stop_index = index - 1
        left_start_index = index
        while index < events_count and events[index].start == current_start:
            index += 1
        left_stop_index = index - 1
        has_right_events = right_stop_index >= right_start_index
        has_left_events = left_stop_index >= left_start_index
        if has_right_events:
            result[right_start_index:right_stop_index] = range(
                    right_start_index + 1, right_stop_index + 1)
            result[right_stop_index] = (left_stop_index
                                        if has_left_events
                                        else right_start_index)
        if has_left_events:
            result[left_start_index] = (right_start_index
                                        if has_right_events
                                        else left_stop_index)
            result[left_start_index + 1:left_stop_index + 1] = range(
                    left_start_index, left_stop_index)
    return result
